# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import contextlib
import warnings
import hashlib
import stat
import os
import io
import collections
import uuid as _uuid
import yaml
import zipfile
import pathlib
import shutil
import subprocess

import decorator

READ_ONLY_FILE = stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH
READ_ONLY_DIR = stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH | stat.S_IRUSR \
    | stat.S_IRGRP | stat.S_IROTH
USER_GROUP_RWX = stat.S_IRWXU | stat.S_IRWXG
OTHER_NO_WRITE = stat.S_IRWXU | stat.S_IRWXG | stat.S_IROTH | stat.S_IXOTH


def get_view_name(view):
    from .format import FormatBase
    if not isinstance(view, type):
        view = view.__class__

    if issubclass(view, FormatBase):
        # Not qualname because we don't have a notion of "nested" formats
        return view.__name__

    return ':'.join([view.__module__, view.__qualname__])


def tuplize(x):
    if type(x) is not tuple:
        return (x,)
    return x


def overrides(cls):
    def decorator(func):
        if not hasattr(cls, func.__name__):
            raise AssertionError("%r does not override %r"
                                 % (func, cls.__name__))
        return func
    return decorator


def superscript(number):
    table = {
        '0': chr(8304), '1': chr(185), '2': chr(178), '3': chr(179),
        **{str(i): chr(x) for i, x in enumerate(range(8308, 8314), 4)},
        'a': chr(7491), 'e': chr(7497), 'f': chr(7584), 'i': chr(8305),
        'n': chr(8319), '-': chr(8315), '.': chr(39), ',': chr(39)
    }
    return ''.join([table[d] for d in str(number)])


def find_duplicates(iterable):
    """Find duplicate values in an iterable.

    Parameters
    ----------
    iterable : iterable
        Iterable to search for duplicates.

    Returns
    -------
    set
        Values that are duplicated in `iterable`.

    Notes
    -----
    Values in `iterable` must be hashable.

    """
    # Modified from https://stackoverflow.com/a/9835819/3776794 to return
    # duplicates instead of remove duplicates from an iterable.
    seen = set()
    duplicates = set()
    for value in iterable:
        if value in seen:
            duplicates.add(value)
        else:
            seen.add(value)
    return duplicates


# Concept from: http://stackoverflow.com/a/11157649/579416
def duration_time(relative_delta):
    attrs = ['years', 'months', 'days', 'hours', 'minutes', 'seconds',
             'microseconds']
    results = []
    for attr in attrs:
        value = getattr(relative_delta, attr)
        if value != 0:
            if value == 1:
                # Remove plural 's'
                attr = attr[:-1]
            results.append("%d %s" % (value, attr))
    if results:
        text = results[-1]
        if results[:-1]:
            text = ', and '.join([', '.join(results[:-1]), text])
        return text
    else:
        # Great Scott! No time has passed!
        return '0 %s' % attrs[-1]


def has_md5sum_native():
    return shutil.which('md5sum') is not None


def md5sum(filepath):
    if has_md5sum_native():
        return md5sum_native(filepath)
    else:
        return md5sum_python(filepath)


def md5sum_python(filepath):
    md5 = hashlib.md5()
    with open(str(filepath), mode='rb') as fh:
        for chunk in iter(lambda: fh.read(io.DEFAULT_BUFFER_SIZE), b""):
            md5.update(chunk)
    return md5.hexdigest()


def md5sum_native(filepath):
    result = subprocess.run(['md5sum', str(filepath)],
                            check=True, capture_output=True, text=True)
    _, digest = from_checksum_format(result.stdout)
    return digest


def md5sum_zip(zf: zipfile.ZipFile, filepath: str) -> str:
    """
    Given a ZipFile object and relative filepath within the zip archive,
    returns the md5sum of the file
    """
    md5 = hashlib.md5()
    with zf.open(filepath) as fh:
        for chunk in iter(lambda: fh.read(io.DEFAULT_BUFFER_SIZE), b""):
            md5.update(chunk)
    return md5.hexdigest()


def md5sum_directory(directory):
    if has_md5sum_native():
        md5sum = md5sum_native
    else:
        md5sum = md5sum_python

    directory = str(directory)
    sums = collections.OrderedDict()
    for root, dirs, files in os.walk(directory, topdown=True):
        dirs[:] = sorted([d for d in dirs if not d[0] == '.'])
        for file in sorted(files):
            if file[0] == '.':
                continue

            path = os.path.join(root, file)
            sums[os.path.relpath(path, start=directory)] = md5sum(path)
    return sums


def md5sum_directory_zip(zf: zipfile.ZipFile) -> dict:
    """
    Returns a mapping of fp/checksum pairs for all files in zf.

    The root dir has been removed from these filepaths. This mimics the output
    in checksums.md5 (without sorted descent), but is not generalizable beyond
    QIIME 2 archives.
    """
    sums = dict()
    for file in zf.namelist():
        fp = pathlib.Path(file)
        if fp.name != 'checksums.md5':
            file_parts = list(fp.parts)
            fp_w_o_root_uuid = pathlib.Path(*(file_parts[1:]))
            sums[str(fp_w_o_root_uuid)] = md5sum_zip(zf, file)
    return sums


def to_checksum_format(filepath, checksum):
    # see https://www.gnu.org
    # /software/coreutils/manual/html_node/md5sum-invocation.html
    if '\\' in filepath or '\n' in filepath:
        filepath = filepath.replace('\\', '\\\\').replace('\n', '\\n')
        checksum = '\\' + checksum

    return '%s  %s' % (checksum, filepath)


def from_checksum_format(line):
    line = line.rstrip('\n')
    parts = line.split('  ', 1)
    if len(parts) < 2:
        parts = line.split(' *', 1)

    checksum, filepath = parts

    if checksum[0] == '\\':
        chars = ''
        escape = False
        # Gross, but regular `.replace` will overlap with itself and
        # negative lookbehind in regex is *probably* harder than scanning
        for char in filepath:
            # 1) Escape next character
            if not escape and char == '\\':
                escape = True
                continue

            # 2) Handle escape sequence
            if escape:
                try:
                    chars += {'\\': '\\', 'n': '\n'}[char]
                except KeyError:
                    chars += '\\' + char  # Wasn't an escape after all
                escape = False
                continue

            # 3) Nothing interesting
            chars += char

        checksum = checksum[1:]
        filepath = chars

    return filepath, checksum


@contextlib.contextmanager
def warning():
    def _warnformat(msg, category, filename, lineno, file=None, line=None):
        return '%s:%s: %s: %s\n' % (filename, lineno, category.__name__, msg)

    default_warn_format = warnings.formatwarning
    try:
        warnings.formatwarning = _warnformat
        warnings.filterwarnings('always')
        yield warnings.warn
    finally:
        warnings.formatwarning = default_warn_format


# Descriptor protocol for creating an attribute that is bound to an
# (arbitrarily nested) attribute accessible to the instance at runtime.
class LateBindingAttribute:
    def __init__(self, attribute):
        self._attribute = attribute

    def __get__(self, obj, cls=None):
        attrs = self._attribute.split('.')
        curr_attr = obj
        for attr in attrs:
            curr_attr = getattr(curr_attr, attr)
        return staticmethod(curr_attr).__get__(obj, cls)


# Removes the first parameter from a callable's signature.
class DropFirstParameter(decorator.FunctionMaker):
    @classmethod
    def from_function(cls, function):
        return cls.create(function, "return None", {})

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.signature = self._remove_first_arg(self.signature)
        self.shortsignature = self._remove_first_arg(self.shortsignature)

    def _remove_first_arg(self, string):
        return ",".join(string.split(',')[1:])[1:]


def _immutable_error(obj, *args):
    raise TypeError('%s is immutable.' % obj.__class__.__name__)


class ImmutableBase:
    def _freeze_(self):
        """Disables __setattr__ when called. It is idempotent."""
        self._frozen = True  # The particular value doesn't matter

    __delattr__ = __setitem__ = __delitem__ = _immutable_error

    def __setattr__(self, *args):
        # This doesn't stop silly things like
        # object.__setattr__(obj, ...), but that's a pretty rude thing
        # to do anyways. We are just trying to avoid accidental mutation.
        if hasattr(self, '_frozen'):
            _immutable_error(self)
        super().__setattr__(*args)


def sorted_poset(iterable, *, key=None, reverse=False):
    values = list(iterable)
    elements = values
    if key is not None:
        elements = [key(x) for x in values]

    result = []
    sorted_elements = []
    for value, element in zip(values, elements):
        idx = 0
        for idx, placed in enumerate(sorted_elements, 1):
            if element <= placed:
                idx -= 1
                break

        result.insert(idx, value)
        sorted_elements.insert(idx, element)
    if reverse:
        result = list(reversed(result))
    return result


def is_uuid4(uuid_str):
    # Adapted from https://gist.github.com/ShawnMilo/7777304
    try:
        uuid = _uuid.UUID(hex=uuid_str, version=4)
    except ValueError:
        # The string is not a valid hex code for a UUID.
        return False

    # If uuid_str is a valid hex code, but an invalid uuid4, UUID.__init__
    # will convert it to a valid uuid4.
    return str(uuid) == uuid_str


def set_permissions(path, file_permissions=None, dir_permissions=None,
                    skip_root=False):
    """Set permissions on all directories and files under and including path
    """
    # Panfs is currently causing issues for us setting permissions. We still
    # want to set rwx for user and group before we remove things to ensure we
    # can remove them, but we want to temporarily no-op other permission
    # changes
    if file_permissions != USER_GROUP_RWX:
        file_permissions = None

    if dir_permissions != USER_GROUP_RWX:
        dir_permissions = None

    # Just get out if we aren't doing anything
    if file_permissions is None and dir_permissions is None:
        return

    for directory, _, files in os.walk(path):
        # We may want to set permissions under a directory but not on the
        # directory itself.
        if dir_permissions and not (skip_root and directory == str(path)):
            try:
                os.chmod(directory, dir_permissions)
            except FileNotFoundError:
                pass

        for file in files:
            if file_permissions:
                try:
                    os.chmod(os.path.join(directory, file), file_permissions)
                except FileNotFoundError:
                    pass


def touch_under_path(path):
    """Touches everything under a given path to ensure they don't get culled by
    Mac
    """
    for directory, _, files in os.walk(path):
        try:
            os.utime(directory, None, follow_symlinks=False)
        except FileNotFoundError:
            pass

        for file in files:
            try:
                os.utime(
                    os.path.join(directory, file), None, follow_symlinks=False)
            except FileNotFoundError:
                pass


def load_action_yaml(path):
    """Takes a path to an unzipped Aritfact and loads its action.yaml with
    yaml.safe_load
    """
    # TODO: Make these actually do something useful at least for the tags
    # that are relevant to what we need out of provenance (this is partially
    # done)
    def ref_constructor(loader, node):
        # We only care about the name of the thing we are referencing which
        # is at the end of this list
        return node.value.split(':')[-1]

    def cite_constructor(loader, node):
        return node.value

    def metadata_constructor(loader, node):
        # Use the md5sum of the metadata as its identifier, so we can tell
        # if two artifacts used the same metadata input
        metadata_path = prov_path / node.value
        return md5sum(metadata_path)

    yaml.constructor.SafeConstructor.add_constructor('!ref', ref_constructor)
    yaml.constructor.SafeConstructor.add_constructor('!cite', cite_constructor)
    yaml.constructor.SafeConstructor.add_constructor(
        '!metadata', metadata_constructor)

    prov_path = path / 'provenance' / 'action'
    action_path = prov_path / 'action.yaml'

    with open(action_path) as fh:
        prov = yaml.safe_load(fh)

    return prov


def create_collection_name(*, name, key, idx, size):
    """ Only accepts kwargs. Creates a name for a collection item in a
        standardized way. Assumes 0 based indexing.
    """
    return [name, key, f'{idx + 1}/{size}']



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import re
from pkg_resources import iter_entry_points
from typing import Dict, TYPE_CHECKING

import qiime2.sdk
import qiime2.core.type as qtype
import qiime2.core.type.parse as _parse
from qiime2.core.type import (
    is_semantic_type, is_primitive_type, is_collection_type, is_metadata_type,
    is_visualization_type, interrogate_collection_type, parse_primitive,
    is_union, is_metadata_column_type, is_parallel_type)

if TYPE_CHECKING:
    from qiime2.sdk.usage import UsageDriver

__all__ = [
    'is_semantic_type', 'is_primitive_type', 'is_collection_type',
    'is_metadata_type', 'is_visualization_type', 'interrogate_collection_type',
    'type_from_ast', 'parse_primitive', 'parse_type', 'parse_format',
    'actions_by_input_type', 'is_union', 'is_metadata_column_type',
    'is_parallel_type'
]


def camel_to_snake(name: str) -> str:
    """
    There are more comprehensive and faster ways of doing this (incl compiling)
    but it handles acronyms in semantic types nicely
    e.g. EMPSingleEndSequences -> emp_single_end_sequences
    c/o https://stackoverflow.com/a/1176023/9872253
    """
    # this will frequently be called on QIIME type expressions, so drop [ and ]
    name = re.sub(r'[\[\]]', '', name)
    # camel to snake
    name = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', name).lower()


def type_from_ast(ast, scope=None):
    """Convert a type ast (from `.to_ast()`) to a type expression.

    Parameters
    ----------
    ast : json compatible object
        The abstract syntax tree produced by `to_ast` on a type.
    scope : dict
        A dictionary to use between multiple calls to share scope between
        different types. This allows type variables from the same type map
        to be constructed from an equivalent type map. Scope should be shared
        within a given call signature, but not between call signatures.

    Returns
    -------
    type expression

    """
    return _parse.ast_to_type(ast, scope=scope)


def parse_type(string, expect=None):
    """Convert a string into a type expression

    Parameters
    ----------
    string : str
        The string type expression to convert into a TypeExpression
    expect : {'semantic', 'primitive', 'visualization'}, optional
        Will raise a TypeError if the resulting TypeExpression is not a member
        of `expect`.

    Returns
    -------
    type expression

    """
    if expect is not None and expect not in {'semantic', 'primitive',
                                             'visualization'}:
        raise ValueError("`expect` got %r, must be 'semantic', 'primitive',"
                         " 'visualization', or None." % (expect,))

    type_expr = _parse.ast_to_type(_parse.string_to_ast(string))

    if expect is None:
        pass
    elif expect == 'semantic' and qtype.is_semantic_type(type_expr):
        pass
    elif expect == 'primitive' and qtype.is_primitive_type(type_expr):
        pass
    elif expect == 'visualization' and type_expr == qtype.Visualization:
        pass
    else:
        raise TypeError("Type expression %r is not a %s type."
                        % (type_expr, expect))
    return type_expr


def parse_format(format_str):
    if format_str is None:
        return None

    pm = qiime2.sdk.PluginManager()
    try:
        format_record = pm.formats[format_str]
    except KeyError:
        raise TypeError("No format: %s" % format_str)
    return format_record.format


def actions_by_input_type(string):
    """Plugins and actions that have as input the artifact type (string)

    Parameters
    ----------
    string : str
        QIIME2 artifact type

    Returns
    -------
    list of tuples: [(q2.plugin, [q2.actions, ...]), ...]
    """
    commands = []
    if string is not None:
        query_type = qiime2.sdk.util.parse_type(string)

        pm = qiime2.sdk.PluginManager()
        for pgn, pg in pm.plugins.items():
            actions = list({a for an, a in pg.actions.items()
                            for iname, i in a.signature.inputs.items()
                            if i.qiime_type >= query_type})
            if actions:
                commands.append((pg, actions))

    return commands


def validate_result_collection_keys(*args):
    """Validate one or more strings intended for use as ResultCollection keys.

    This can be called on one or more keys provided as arguments:
    qiime2.sdk.util.validate_result_collection_keys('@', 'a1@')

    Or on a list, by unpacking it in the call:
    l = ['@', 'a1@']
    qiime2.sdk.util.validate_result_collection_keys(*l)
    """
    invalid_keys = []
    for key in args:
        if not isinstance(key, str) or bool(re.search(r'[^\w+-.]', key)):
            invalid_keys.append(key)

    if len(invalid_keys) > 0:
        raise KeyError('Invalid key(s) provided for ResultCollection. '
                       'ResultCollection keys must be strings and may only '
                       'contain the following characters: A-Z, a-z, 0-9, +, '
                       '-, ., and _. Offending keys include: '
                       f'{", ".join(map(str, invalid_keys))}')


def view_collection(collection, view_type):
    return {k: v.view(view_type) for k, v in collection.items()}


def get_available_usage_drivers() -> Dict[str, 'UsageDriver']:
    '''
    Discovers all usage drivers registered under the entry point group
    'qiime2.usage_drivers'.

    Returns
    -------
    dict of entry point name -> UsageDriver
        Where keys are the entry point names as registered in setup.py and the
        values are types of the available usage drivers (not instances
        themselves).
    '''
    return {
        entry_point.name: entry_point.resolve() for entry_point in
        iter_entry_points(group='qiime2.usage_drivers')
    }



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------
import codecs
import pathlib
import re
import warnings

from typing import Tuple
from zipfile import ZipFile


def get_root_uuid(zf: ZipFile) -> str:
    '''
    Returns the root UUID of a QIIME 2 Archive.

    Parameters
    ----------
    zf : ZipFile
        The zipfile object of an archive.

    Returns
    -------
    str
        The uuid of the root artifact in the archive.
    '''
    return pathlib.Path(zf.namelist()[0]).parts[0]


def get_nonroot_uuid(fp: pathlib.Path) -> str:
    '''
    For non-root provenance files, get the Result's uuid from its path.

    Parameters
    ----------
    fp : pathlib.Path
        The path to a file in a non-root artifact inside an archive, relative
        to archive root.

    Returns
    -------
    str
        The uuid of the non-root artifact.
    '''
    if fp.name == 'action.yaml':
        return fp.parts[-3]
    return fp.parts[-2]


_VERSION_MATCHER = (
    r'QIIME 2\n'
    r'archive: [0-9]{1,2}$\n'
    r'framework: '
    r'(?:20[0-9]{2}|2)\.(?:[1-9][0-2]?|0)\.[0-9](?:\.dev[0-9]?)?'
    r'(?:\+[.\w]+)?\Z'
)


def parse_version(zf: ZipFile) -> Tuple[str, str]:
    '''
    Finds and parses the VERSION file inside of an archive.

    Parameters
    ----------
    zf : ZipFile
        The zipfile object of an archive.

    Returns
    -------
    tuple of (str, str)
        The archive version and framework version of the archive.
    '''
    uuid = get_root_uuid(zf)
    version_fp = pathlib.Path(uuid) / 'VERSION'

    try:
        with zf.open(str(version_fp)) as v_fp:
            version_contents = str(v_fp.read().strip(), 'utf-8')
    except KeyError:
        raise ValueError(
            f'Malformed Archive: VERSION file for node {uuid} misplaced '
            f'or nonexistent.\nArchive {zf.filename} may be corrupt or '
            'provenance may be false.'
        )

    if not re.match(_VERSION_MATCHER, version_contents, re.MULTILINE):
        warnings.filterwarnings(
            'ignore', 'invalid escape sequence', DeprecationWarning
        )
        version_match_repr = codecs.decode(
            _VERSION_MATCHER.encode('utf-8'), 'unicode-escape'
        )
        raise ValueError(
            f'Malformed Archive: VERSION file out of spec in {zf.filename}.\n'
            f'Should match this regular expression:\n{version_match_repr}\n'
            f'Actually looks like:\n{version_contents}\n'
        )

    _, archive_version, framework_version = [
        line.strip().split()[-1] for line in
        version_contents.split(sep='\n') if line
    ]
    return (archive_version, framework_version)



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------
from __future__ import annotations
import copy
from typing import Any, List, Optional, Set

import networkx as nx
import os
from pathlib import Path
from networkx.classes.reportviews import NodeView

from ._checksum_validator import ValidationCode, ChecksumDiff
from .archive_parser import (
    Config, ParserResults, ProvNode, Parser, ArchiveParser
)


class ProvDAG:
    '''
    A directed acyclic graph (DAG) representing the provenance of one or more
    QIIME 2 Archives (.qza or .qzv files).

    Parameters
    ----------
    artifact_data : Any
        The input data payload for the ProvDAG. Often the path to a file or
        directory on disk.
    validate_checksums : bool
        If True, Archives will be validated against their checksums.md5
        manifests.
    parse_metadata : bool
        If True, the metadata captured in the input Archives will be parsed
        and included in the ProvDAG.
    recurse : bool
        If True, and if artifact_data is a directory, will recursively parse
        all .qza and .qzv files within subdirectories.
    verbose : bool
        If True, will print parsed filenames to stdout, indicating progress.

    Attributes
    ----------
    dag : nx.DiGraph
        A Directed Acyclic Graph (DAG) representing the complete provenance
        of one or more QIIME 2 Artifacts. This DAG is comprehensive, including
        pipeline "alias" nodes as well as the inner nodes that compose each
        pipeline.
    parsed_artifact_uuids : Set[UUID]
        The set of user-passed terminal node uuids. Used to generate properties
        like `terminal_uuids`, this is a superset of terminal_uuids.
    terminal_uuids : Set[UUID]
        The set of terminal node ids present in the DAG, not including inner
        pipeline nodes.
    terminal_nodes : Set[ProvNode]
        The terminal ProvNodes present in the DAG, not including inner pipeline
        nodes.
    provenance_is_valid : ValidationCode
        The canonical indicator of provenance validity for a dag, this contains
        the lowest ValidationCode from all parsed Artifacts unioned into a
        given ProvDAG.
    checksum_diff : ChecksumDiff
        A ChecksumDiff representing all added, removed, and changed filepaths
        from all parsed Artifacts. If an artifact's checksums.md5 file is
        missing, this may be None. When multiple artifacts are unioned, this
        field prefers ChecksumDiffs over Nonetypes, which will be dropped.
    '''
    def __init__(
        self,
        artifact_data: Any = None,
        validate_checksums: bool = True,
        parse_metadata: bool = True,
        recurse: bool = False,
        verbose: bool = False,
    ):
        '''
        Creates a digraph by getting a parser from the parser dispatcher then
        parses the incoming data into a ParserResults, and then loads those
        results into key fields.
        '''
        cfg = Config(validate_checksums, parse_metadata, recurse, verbose)
        parser_results = parse_provenance(cfg, artifact_data)

        self.cfg = cfg
        self._parsed_artifact_uuids = parser_results.parsed_artifact_uuids
        self.dag = parser_results.prov_digraph
        self._provenance_is_valid = parser_results.provenance_is_valid
        self._checksum_diff = parser_results.checksum_diff

        # clear cache whenever we create a new ProvDAG
        self._terminal_uuids = None

    def __repr__(self) -> str:
        return (
            'ProvDAG representing the provenance of the Artifacts: '
            f'{self._parsed_artifact_uuids}'
        )

    __str__ = __repr__

    def __len__(self) -> int:
        return len(self.dag)

    def __eq__(self, other) -> bool:
        if self.__class__ != other.__class__:
            return False
        if not nx.is_isomorphic(self.dag, other.dag):
            return False
        return True

    def __iter__(self):
        return iter(self.dag)

    @property
    def terminal_uuids(self) -> Set[str]:
        '''
        The UUIDs of the terminal nodes in the DAG, generated by selecting all
        nodes in a collapsed view of self.dag with an out-degree of zero.

        We memoize the set of terminal UUIDs to prevent unnecessary traversals,
        so must set self._terminal_uuid back to None in any method that
        modifies the structure of self.dag, or the nodes themselves (which are
        literal UUIDs). These methods include self.union() and
        self.relabel_nodes().
        '''
        if self._terminal_uuids is not None:
            return self._terminal_uuids
        cv = self.collapsed_view
        self._terminal_uuids = {
            uuid for uuid, out_degree in cv.out_degree()
            if out_degree == 0
        }
        return self._terminal_uuids

    @property
    def parsed_artifact_uuids(self) -> Set[str]:
        '''
        The set of user-passed terminal node uuids. Used to generate properties
        like self.terminal_uuids.
        '''
        return self._parsed_artifact_uuids

    @property
    def terminal_nodes(self) -> Set[ProvNode]:
        '''The terminal ProvNodes in the DAG's provenance.'''
        return {self.get_node_data(uuid) for uuid in self.terminal_uuids}

    @property
    def provenance_is_valid(self) -> ValidationCode:
        return self._provenance_is_valid

    @property
    def checksum_diff(self) -> Optional[ChecksumDiff]:
        return self._checksum_diff

    @property
    def nodes(self) -> NodeView:
        return self.dag.nodes

    @property
    def collapsed_view(self) -> nx.DiGraph:
        '''
        Returns a subsetted graphview of self.dag containing only nodes
        that are exterior to pipeline actions.
        '''
        outer_nodes = set()
        for terminal_uuid in self._parsed_artifact_uuids:
            outer_nodes |= self.get_outer_provenance_nodes(terminal_uuid)

        return nx.subgraph_view(self.dag, lambda node: node in outer_nodes)

    def has_edge(self, start_node: str, end_node: str) -> bool:
        return self.dag.has_edge(start_node, end_node)

    def node_has_provenance(self, uuid: str) -> bool:
        return self.dag.nodes[uuid]['has_provenance']

    def get_node_data(self, uuid: str) -> ProvNode:
        '''Returns a ProvNode from this ProvDAG selected by UUID.'''
        return self.dag.nodes[uuid]['node_data']

    def predecessors(self, node: str, dag: nx.DiGraph = None) -> Set[str]:
        '''
        Returns the parent UUIDs of a given node.

        Parameters
        ----------
        node : str
            The uuid of the node of interest.
        dag : nx.DiGraph
            The current provenance graph.

        Returns
        -------
        set of str
            The uuids of the parents of the node.
        '''
        dag = self.collapsed_view if dag is None else dag
        return set(self.dag.predecessors(node))

    @classmethod
    def union(cls, dags: List[ProvDAG]) -> ProvDAG:
        '''
        Class method that creates a new ProvDAG by unioning two or more graphs.

        The returned ProvDAG._parsed_artifact_uuids will include uuids from all
        dags, and other DAG attributes are reduced conservatively.

        Parameters
        ----------
        dags : list of ProvDAG
            A collection of ProvDAGs to union.

        Returns
        -------
        ProvDAG
            A single ProvDAG representing the union of all inputs.
        '''
        if len(dags) < 2:
            raise ValueError('Please pass at least two ProvDAGs.')

        union_dag = ProvDAG()
        union_dag.dag = nx.compose_all((dag.dag for dag in dags))

        union_dag._parsed_artifact_uuids = dags[0]._parsed_artifact_uuids
        union_dag._provenance_is_valid = dags[0]._provenance_is_valid
        union_dag._checksum_diff = dags[0].checksum_diff

        for next_dag in dags[1:]:
            union_dag._parsed_artifact_uuids = \
                union_dag._parsed_artifact_uuids.\
                union(next_dag._parsed_artifact_uuids)

            union_dag._provenance_is_valid = min(
                union_dag.provenance_is_valid, next_dag.provenance_is_valid
            )

            union_dag.cfg.parse_study_metadata = min(
                union_dag.cfg.parse_study_metadata,
                next_dag.cfg.parse_study_metadata
            )
            union_dag.cfg.perform_checksum_validation = min(
                union_dag.cfg.perform_checksum_validation,
                next_dag.cfg.perform_checksum_validation
            )

            # Here we retain as much data as possible, preferencing any
            # ChecksumDiff over None. This might mean we keep an empty
            # ChecksumDiff and drop a None ChecksumDiff, the latter of which is
            # used to indicate a missing checksums.md5 file in v5+ archives.
            # union_dag._provenance_is_valid will still be INVALID however.
            if next_dag.checksum_diff is None:
                continue

            if union_dag.checksum_diff is None:
                union_dag._checksum_diff = next_dag.checksum_diff
            else:
                union_dag.checksum_diff.added.update(
                    next_dag.checksum_diff.added
                )
                union_dag.checksum_diff.removed.update(
                    next_dag.checksum_diff.removed
                )
                union_dag.checksum_diff.changed.update(
                    next_dag.checksum_diff.changed
                )

        # make union._terminal_uuids be recalculated on next access
        union_dag._terminal_uuids = None
        return union_dag

    def get_outer_provenance_nodes(self, node_id: str = None) -> Set[str]:
        '''
        Performs depth-first traversal of a node's ancestors. Skips over nodes
        that are interior to a pipeline because pipeline output nodes point to
        the pipeline's inputs as parents not their direct parents inside
        the pipeline.

        Parameters
        ----------
        node_id : str
            The uuid of the node for which to discover ancestors.

        Returns
        -------
        set of str
            All ancestor uuids according to the above definition.
        '''
        nodes = set() if node_id is None else {node_id}
        parents = [edge_pair[0] for edge_pair in self.dag.in_edges(node_id)]
        for uuid in parents:
            nodes = nodes | self.get_outer_provenance_nodes(uuid)
        return nodes


# TODO: can this get nuked?
class EmptyParser(Parser):
    '''
    Creates empty ProvDAGs.
    Disregards Config, because it's not meaningful in this context.
    '''
    accepted_data_types = 'None'

    @classmethod
    def get_parser(cls, artifact_data: Any) -> Parser:
        if artifact_data is None:
            return EmptyParser()
        else:
            raise TypeError(f' in EmptyParser: {artifact_data} is not None.')

    def parse_prov(self, cfg: Config, data: None) -> ParserResults:
        '''
        Returns a static ParserResults with empty parsed_artifact_uuids,
        an empty graph, a valid ValidationCode, and a None ChecksumDiff.
        '''
        return ParserResults(
            parsed_artifact_uuids=set(),
            prov_digraph=nx.DiGraph(),
            provenance_is_valid=ValidationCode.VALID,
            checksum_diff=None,
        )


class DirectoryParser(Parser):
    accepted_data_types = \
        'filepath to a directory containing .qza/.qzv archives'

    @classmethod
    def get_parser(cls, artifact_data: Any) -> Parser:
        '''
        Return a DirectoryParser if appropriate.

        Parameters
        ----------
        artifact_data : Any
            Ideally a path to a directory containing one or more archives, but
            may be a different type during searches for other Parsers.

        Raises
        ------
        TypeError
            If something other than a str or path-like object is input.
        ValueError
            If the path does not point to a directory.
        '''
        try:
            is_dir = os.path.isdir(artifact_data)
        except TypeError:
            t = type(artifact_data)
            raise TypeError(
                f' in DirectoryParser: expects a directory, not a {t}.'
            )

        if not is_dir:
            raise ValueError(
                f' in DirectoryParser: {artifact_data} '
                'is not a valid directory.'
            )

        return DirectoryParser()

    def parse_prov(self, cfg: Config, data: str) -> ParserResults:
        '''
        Iterates over the directory's .qza and .qzv files, parsing them if
        their terminal node isn't already in the DAG.

        This behavior assumes that the ArchiveParsers capture all nodes within
        the archives they parse by default.

        Parameters
        ----------
        cfg : Config
            User-selected configuration options for whether to perform checksum
            validation, whether to parse study metadata, whether to recursively
            parse nested directories of artifacts, and whether to print status
            messages during processing.
        data : Any
            The path to the directory containing artifacts.

        Returns
        -------
        ParserResults
            A dataclass that stores the parsed artifact uuids, the parsed
            networkx graph, the provenance-is-valid flag, and the
            checksum diff.
        '''
        dir_name = Path(str(data).rstrip('/') + os.sep)
        if cfg.recurse:
            artifacts_to_parse = list(dir_name.rglob('*.qz[av]'))
            err_msg = (
                f'No .qza or .qzv files present in {dir_name} or any '
                'directory nested within it.'
            )
        else:
            artifacts_to_parse = list(dir_name.glob('*.qz[av]'))
            err_msg = (
                f'No .qza or .qzv files present in {dir_name}. Did you '
                'mean to recurse into nested directories?'
            )
        if not artifacts_to_parse:
            raise ValueError(err_msg)

        dag = ProvDAG()
        for archive in artifacts_to_parse:
            if cfg.verbose:
                print("parsing", archive)
            dag = ProvDAG.union([
                dag,
                ProvDAG(
                    archive,
                    cfg.perform_checksum_validation,
                    cfg.parse_study_metadata
                )
            ])

        return ParserResults(
            dag._parsed_artifact_uuids,
            dag.dag,
            dag.provenance_is_valid,
            dag.checksum_diff,
        )


def archive_not_parsed(root_uuid: str, dag: ProvDAG) -> bool:
    '''
    Checks if the archive with root_uuid has not already been parsed into dag
    which is defined as either not in the dag at all, or added only as a
    !no-provenance parent uuid.

    Parameters
    ----------
    root_uuid : str
        The root uuid of an archive.
    dag : ProvDAG
        The ProvDAG in which to search for the root uuid.

    Returns
    -------
    bool
        Indicating whether the archive represented by root_uuid has not been
        parsed into the ProvDAG.
    '''
    if root_uuid not in dag.dag:
        return True
    elif dag.get_node_data(root_uuid) is None:
        return True
    return False


class ProvDAGParser(Parser):
    '''
    Effectively a ProvDAG copy constructor, this "parses" a ProvDAG, loading
    its data into a new ProvDAG.

    Disregards Config, because it's not meaningful in this context.
    '''
    accepted_data_types = 'ProvDAG'

    @classmethod
    def get_parser(cls, artifact_data: Any) -> Parser:
        '''
        Returns ProvDAGParser if appropriate.

        Parameters
        ----------
        artifact_data : Any
            Hopefully a ProvDAG but may be a different type during searches for
            the proper Parser.

        Returns
        -------
        ProvDAGParser
            An instance of ProvDAGParser if artifact_data is a ProvDAG.

        Raises
        ------
        TypeError
            If artifact_data is not a ProvDAG.
        '''
        if isinstance(artifact_data, ProvDAG):
            return ProvDAGParser()
        else:
            raise TypeError(
                f' in ProvDAGParser: {artifact_data} is not a ProvDAG.'
            )

    def parse_prov(self, cfg: Config, dag: ProvDAG) -> ParserResults:
        '''
        Parses a ProvDAG returning a ParserResults by deep copying existing
        attributes that live on the ProvDAG and make up a ParserResults.

        Parameters
        ----------
        cfg : Config
            Ignored because a ProvDAG is not being constructed from scratch.
            Present for inheritance purposes.
        dag : ProvDAG
            The ProvDAG to parse, read: copy attributes from.

        Returns
        -------
        ParserResults
            A dataclass that stores the parsed artifact uuids, the parsed
            networkx graph, the provenance-is-valid flag, and the
            checksum diff.
        '''
        return ParserResults(
            copy.deepcopy(dag._parsed_artifact_uuids),
            copy.deepcopy(dag.dag),
            copy.deepcopy(dag.provenance_is_valid),
            copy.deepcopy(dag.checksum_diff),
        )


def parse_provenance(cfg: Config, payload: Any) -> ParserResults:
    '''
    Parses some data payload into a ParserResults object ingestible by ProvDAG.

    Parameters
    ----------
    cfg : Config
        A dataclass that stores four boolean flags: whether to perform
        checksum validation, whether to parse study metadata, whether to
        recursively parse nested directories, and whether to enable verbose
        mode.
    payload : Any
        The payload to attempt to parse, commonly a path to an archive or
        directory containing archives.

    Returns
    -------
    ParserResults
        A dataclass that stores the parsed artifact uuids, the parsed
        networkx graph, the provenance-is-valid flag, and the
        checksum diff.

    '''
    parser = select_parser(payload)
    return parser.parse_prov(cfg, payload)


def select_parser(payload: Any) -> Parser:
    '''
    Attempts to find a parser that can handle some given payload.

    Parameters
    ----------
    payload : Any
        The payload for which to find a parser.

    Returns
    -------
    Parser
        The appropriate Parser for the payload type.

    Raises
    ------
    UnparseableDataError
        If no appropriate parser could be found for the payload.
    '''
    _PARSER_TYPE_REGISTRY = [
        ArchiveParser,
        DirectoryParser,
        ProvDAGParser,
        EmptyParser
    ]

    accepted_data_types = [
        parser.accepted_data_types for parser in _PARSER_TYPE_REGISTRY
    ]

    optional_parser = None
    errors = []
    for parser in _PARSER_TYPE_REGISTRY:
        try:
            optional_parser = parser.get_parser(payload)
            if optional_parser is not None:
                return optional_parser
        except Exception as e:
            errors.append(e)

    err_msg = (
        f'Input data {payload} is not supported.\n'
        'Parsers are available for the following data types: '
        f'{accepted_data_types}.\n'
        'The following errors were caught while trying to identify a parser '
        'that can_handle this input data:\n'
    )
    for e in errors:
        err_msg += str(type(e)) + str(e) + '\n'
    raise UnparseableDataError(err_msg)


class UnparseableDataError(Exception):
    pass



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------
import codecs
import os
import shutil
import tempfile
import unittest
import zipfile

import qiime2
from qiime2 import Artifact
from qiime2.sdk.plugin_manager import PluginManager
from qiime2.core.archive.archiver import Archiver

from .testing_utilities import (
    write_zip_archive, monkeypatch_archive_version,
    monkeypatch_framework_version
)
from ..util import _VERSION_MATCHER, parse_version


class TestVersionParser(unittest.TestCase):
    def setUp(self):
        self.pm = PluginManager()
        self.dp = self.pm.plugins['dummy-plugin']
        self.tempdir = tempfile.mkdtemp(
            prefix='qiime2-test-version-parser-temp-'
        )
        self.framework_version_exp = qiime2.__version__
        self.archive_version_exp = Archiver.CURRENT_FORMAT_VERSION

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def test_parse_version(self):
        int_seq = Artifact.import_data('IntSequence1', [1, 2, 3])
        int_seq.save(os.path.join(self.tempdir, 'int-seq.qza'))
        fp = os.path.join(self.tempdir, 'int-seq.qza')
        with zipfile.ZipFile(fp) as zf:
            actual = parse_version(zf)
            self.assertEqual(
                actual, (self.archive_version_exp, self.framework_version_exp)
            )

    def test_parse_version_old_archive_format(self):
        archive_version_exp = '2'
        fp = os.path.join(self.tempdir, 'int-seq-av2.qza')

        with monkeypatch_archive_version(archive_version_exp):
            int_seq = Artifact.import_data('IntSequence1', [1, 2, 3])
            int_seq.save(fp)

        with zipfile.ZipFile(fp) as zf:
            actual = parse_version(zf)
            self.assertEqual(
                actual, (archive_version_exp, self.framework_version_exp)
            )

    def test_artifact_with_commit_version(self):
        framework_version_exp = '2022.8.0+29.gb053440'
        fp = os.path.join(self.tempdir, 'int-seq-custom-fv.qza')

        with monkeypatch_framework_version(framework_version_exp):
            int_seq = Artifact.import_data('IntSequence1', [1, 2, 3])
            int_seq.save(fp)

        with zipfile.ZipFile(fp) as zf:
            actual = parse_version(zf)
            self.assertEqual(
                actual, (self.archive_version_exp, framework_version_exp)
            )

    def test_parse_version_no_VERSION_file(self):
        int_seq = Artifact.import_data('IntSequence1', [1, 2, 3])
        fp = os.path.join(self.tempdir, 'int-seq-no-v.qza')
        int_seq.save(fp)

        with tempfile.TemporaryDirectory() as tempdir:
            with zipfile.ZipFile(fp) as zf:
                zf.extractall(tempdir)

            uuid = os.listdir(tempdir)[0]
            os.remove(os.path.join(tempdir, uuid, 'VERSION'))
            write_zip_archive(fp, tempdir)

        with zipfile.ZipFile(fp) as zf:
            with self.assertRaisesRegex(ValueError,
                                        '(?s)VERSION.*nonexistent.*'):
                parse_version(zf)

    def test_parse_version_VERSION_file_missing_archive_field(self):
        int_seq = Artifact.import_data('IntSequence1', [1, 2, 3])
        fp = os.path.join(self.tempdir, 'int-seq-no-af.qza')
        int_seq.save(fp)

        with tempfile.TemporaryDirectory() as tempdir:
            with zipfile.ZipFile(fp) as zf:
                zf.extractall(tempdir)

            uuid = os.listdir(tempdir)[0]
            with open(os.path.join(tempdir, uuid, 'VERSION')) as fh:
                lines = fh.readlines()
                missing_archive_lines = [lines[0], lines[2]]

            with open(os.path.join(tempdir, uuid, 'VERSION'), 'w') as fh:
                for line in missing_archive_lines:
                    fh.write(line)

            write_zip_archive(fp, tempdir)

        with zipfile.ZipFile(fp) as zf:
            with self.assertRaisesRegex(ValueError, 'VERSION.*out of spec.*'):
                parse_version(zf)

    def test_parse_version_VERSION_file_extra_field(self):
        int_seq = Artifact.import_data('IntSequence1', [1, 2, 3])
        fp = os.path.join(self.tempdir, 'int-seq-extra-f.qza')
        int_seq.save(fp)

        with tempfile.TemporaryDirectory() as tempdir:
            with zipfile.ZipFile(fp) as zf:
                zf.extractall(tempdir)

            uuid = os.listdir(tempdir)[0]
            with open(os.path.join(tempdir, uuid, 'VERSION'), 'a') as fh:
                fh.write('fourth line\n')

            write_zip_archive(fp, tempdir)

        with zipfile.ZipFile(fp) as zf:
            with self.assertRaisesRegex(ValueError, 'VERSION.*out of spec.*'):
                parse_version(zf)

    '''
    Tests of the regex match itself below
    '''
    def test_version_too_short(self):
        short = (
            r'QIIME 2\n'
            r'archive: 4'
        )
        self.assertNotRegex(short, _VERSION_MATCHER)

    def test_version_too_long(self):
        long = (
            r'QIIME 2\n'
            r'archive: 4\n'
            r'framework: 2019.8.1.dev0\n'
            r'This line should not be here'
        )
        self.assertNotRegex(long, _VERSION_MATCHER)

    splitvm = codecs.decode(_VERSION_MATCHER.encode('utf-8'),
                            'unicode-escape').split(sep='\n')
    re_l1, re_l2, re_l3 = splitvm

    def test_line1_good(self):
        self.assertRegex('QIIME 2\n', self.re_l1)

    def test_line1_bad(self):
        self.assertNotRegex('SHIMMY 2\n', self.re_l1)

    def test_archive_version_1digit_numeric(self):
        self.assertRegex('archive: 1\n', self.re_l2)

    def test_archive_version_2digit_numeric(self):
        self.assertRegex('archive: 12\n', self.re_l2)

    def test_archive_version_bad(self):
        self.assertNotRegex('agama agama\n', self.re_l2)

    def test_archive_version_3digit_numeric(self):
        self.assertNotRegex('archive: 123\n', self.re_l2)

    def test_archive_version_nonnumeric(self):
        self.assertNotRegex('archive: 1a\n', self.re_l2)

    def test_fmwk_version_good_semver(self):
        self.assertRegex('framework: 2.0.6', self.re_l3)

    def test_fmwk_version_good_semver_dev(self):
        self.assertRegex('framework: 2.0.6.dev0', self.re_l3)

    def test_fmwk_version_good_year_month_patch(self):
        self.assertRegex('framework: 2020.2.0', self.re_l3)

    def test_fmwk_version_good_year_month_patch_2digit_month(self):
        self.assertRegex('framework: 2018.11.0', self.re_l3)

    def test_fmwk_version_good_year_month_patch_dev(self):
        self.assertRegex('framework: 2020.2.0.dev1', self.re_l3)

    def test_fmwk_version_good_ymp_2digit_month_dev(self):
        self.assertRegex('framework: 2020.11.0.dev0', self.re_l3)

    def test_fmwk_version_invalid_month(self):
        self.assertNotRegex('framework: 2020.13.0', self.re_l3)

    def test_fmwk_version_invalid_month_leading_zero(self):
        self.assertNotRegex('framework: 2020.03.0', self.re_l3)

    def test_fmwk_version_invalid_year(self):
        self.assertNotRegex('framework: 1953.3.0', self.re_l3)



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from qiime2.core.type.util import is_collection_type
from qiime2.core.type import HashableInvocation
from qiime2.core.cache import get_cache
import qiime2.sdk
from qiime2.sdk.parallel_config import PARALLEL_CONFIG


class Context:
    def __init__(self, parent=None, parallel=False):
        if parent is not None:
            self.action_executor_mapping = parent.action_executor_mapping
            self.executor_name_type_mapping = parent.executor_name_type_mapping
            self.parallel = parent.parallel
            self.cache = parent.cache
        else:
            self.action_executor_mapping = \
                PARALLEL_CONFIG.action_executor_mapping
            # Cast type to str so yaml doesn't think it needs tpo instantiate
            # an executor object when we write this to then read this from
            # provenance
            self.executor_name_type_mapping = \
                None if PARALLEL_CONFIG.parallel_config is None \
                else {v.label: v.__class__.__name__
                      for v in PARALLEL_CONFIG.parallel_config.executors}
            self.parallel = parallel
            self.cache = get_cache()
            # Only ever do this on the root context. We only want to index the
            # pool once before we start adding our own stuff to it.
            if self.cache.named_pool is not None:
                self.cache.named_pool.create_index()

        self._parent = parent
        self._scope = None

    def get_action(self, plugin: str, action: str):
        """Return a function matching the callable API of an action.
        This function is aware of the pipeline context and manages its own
        cleanup as appropriate.
        """
        plugin = plugin.replace('_', '-')
        plugin_action = plugin + ':' + action

        pm = qiime2.sdk.PluginManager()
        try:
            plugin_obj = pm.plugins[plugin]
        except KeyError:
            raise ValueError("A plugin named %r could not be found." % plugin)

        try:
            action_obj = plugin_obj.actions[action]
        except KeyError:
            raise ValueError(
                "An action named %r was not found for plugin %r"
                % (action, plugin))

        # We return this callable which determines whether to return cached
        # results or to run the action requested.
        def deferred_action(*args, **kwargs):
            # The function is the first arg, we ditch that
            args = args[1:]

            # If we have a named_pool, we need to check for cached results that
            # we can reuse.
            #
            # We can short circuit our index checking if any of our arguments
            # are proxies because if we got a proxy as an argument, we know it
            # is a new thing we are computing from a prior step in the pipeline
            # and thus will not be cached. We can only have proxies if we are
            # executing with parsl
            if self.cache.named_pool is not None and (not self.parallel or (
                    self.parallel and not self._contains_proxies(
                        *args, **kwargs))):
                collated_inputs = action_obj.signature.collate_inputs(
                    *args, **kwargs)
                callable_args = action_obj.signature.coerce_user_input(
                    **collated_inputs)

                # Make args and kwargs look how they do when we read them out
                # of a .yaml file (list of single value dicts of
                # input_name: value)
                arguments = []
                for k, v in callable_args.items():
                    arguments.append({k: v})

                invocation = HashableInvocation(plugin_action, arguments)
                if invocation in self.cache.named_pool.index:
                    cached_outputs = self.cache.named_pool.index[invocation]
                    loaded_outputs = {}

                    for name, _type in action_obj.signature.outputs.items():
                        if is_collection_type(_type.qiime_type):
                            loaded_collection = qiime2.sdk.ResultCollection()
                            cached_collection = cached_outputs[name]

                            # Get the order we should load collection items in
                            collection_order = list(cached_collection.keys())
                            self._validate_collection(collection_order)
                            collection_order.sort(key=lambda x: x.idx)

                            for elem_info in collection_order:
                                elem = cached_collection[elem_info]
                                loaded_elem = self.cache.named_pool.load(elem)
                                loaded_collection[
                                    elem_info.item_name] = loaded_elem

                            loaded_outputs[name] = loaded_collection
                        else:
                            output = cached_outputs[name]
                            loaded_outputs[name] = \
                                self.cache.named_pool.load(output)

                    return qiime2.sdk.Results(
                        loaded_outputs.keys(), loaded_outputs.values())

            # If we didn't have cached results to reuse, we need to execute the
            # action.
            #
            # These factories will create new Contexts with this context as
            # their parent. This allows scope cleanup to happen recursively. A
            # factory is necessary so that independent applications of the
            # returned callable receive their own Context objects.
            #
            # The parsl factory is a bit more complicated because we need to
            # pass this exact Context along for a while longer until we run a
            # normal _bind in action/_run_parsl_action. Then we create a new
            # Context with this one as its parent inside of the parsl app
            def _bind_parsl_context(ctx):
                def _bind_parsl_args(*args, **kwargs):
                    return action_obj._bind_parsl(ctx, *args, **kwargs)
                return _bind_parsl_args

            if self.parallel:
                return _bind_parsl_context(self)(*args, **kwargs)

            return action_obj._bind(
                lambda: Context(parent=self))(*args, **kwargs)

        deferred_action = action_obj._rewrite_wrapper_signature(
            deferred_action)
        action_obj._set_wrapper_properties(deferred_action)
        return deferred_action

    def _contains_proxies(self, *args, **kwargs):
        """Returns True if any of the args or kwargs are proxies
        """
        return any(isinstance(arg, qiime2.sdk.proxy.Proxy) for arg in args) \
            or any(isinstance(value, qiime2.sdk.proxy.Proxy) for
                   value in kwargs.values())

    def _validate_collection(self, collection_order):
        """Validate that all indexed items in the collection agree on how
        large the collection should be and that we have that many elements.
        """
        assert all([elem.total == collection_order[0].total
                    for elem in collection_order])
        assert len(collection_order) == collection_order[0].total

    def make_artifact(self, type, view, view_type=None):
        """Return a new artifact from a given view.

        This artifact is automatically tracked and cleaned by the pipeline
        context.
        """
        artifact = qiime2.sdk.Artifact.import_data(type, view, view_type)
        # self._scope WILL be defined at this point, as pipelines always enter
        # a scope before deferring to plugin code. (Otherwise cleanup wouldn't
        # happen)
        self._scope.add_reference(artifact)
        return artifact

    def __enter__(self):
        """For internal use only.

        Creates a scope API that can track references that need to be
        destroyed.

        """
        if self._scope is not None:
            # Prevent odd things from happening to lifecycle cleanup
            raise Exception('Cannot enter a context twice.')
        self._scope = Scope(self)
        return self._scope

    def __exit__(self, exc_type, exc_value, exc_tb):
        if exc_type is not None:
            # Something went wrong, teardown everything
            self._scope.destroy()
        else:
            # Everything is fine, just cleanup internal references and pass
            # ownership off to the parent context.
            parent_refs = self._scope.destroy(local_references_only=True)
            if self._parent is not None and self._parent._scope is not None:
                for ref in parent_refs:
                    self._parent._scope.add_reference(ref)


class Scope:
    def __init__(self, ctx):
        self.ctx = ctx
        self._locals = []
        self._parent_locals = []

    def add_reference(self, ref):
        """Add a reference to something destructable that is owned by this
           scope.
        """
        self._locals.append(ref)

    # NOTE: We end up with both the artifact and the pipeline alias of artifact
    # in the named cache in the end. We only have the pipeline alias in the
    # process pool
    def add_parent_reference(self, ref):
        """Add a reference to something destructable that will be owned by the
           parent scope. The reason it needs to be tracked is so that on
           failure, a context can still identify what will (no longer) be
           returned.
        """
        new_ref = self.ctx.cache.process_pool.save(ref)

        if self.ctx.cache.named_pool is not None:
            self.ctx.cache.named_pool.save(new_ref)

        self._parent_locals.append(new_ref)
        self._parent_locals.append(ref)

        # Return an artifact backed by the data in the cache
        return new_ref

    def destroy(self, local_references_only=False):
        """Destroy all references and clear state.

        Parameters
        ----------
        local_references_only : bool
            Whether to destroy references that will belong to the parent scope.
        Returns
        -------
        list
            The list of references that were not destroyed.

        """
        ctx = self.ctx
        local_refs = self._locals
        parent_refs = self._parent_locals

        # Unset instance state, handy to prevent cycles in GC, and also causes
        # catastrophic failure if some invariant is violated.
        del self.ctx
        del self._locals
        del self._parent_locals

        for ref in local_refs:
            ref._destructor()

        if local_references_only:
            return parent_refs

        for ref in parent_refs:
            ref._destructor()

        ctx.cache.garbage_collection()
        return []



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import abc
import concurrent.futures
import inspect
import tempfile
import textwrap

import decorator
import dill
from parsl.app.app import python_app, join_app

import qiime2.sdk
import qiime2.core.type as qtype
import qiime2.core.archive as archive
from qiime2.core.util import (LateBindingAttribute, DropFirstParameter,
                              tuplize, create_collection_name)
from qiime2.sdk.proxy import Proxy


def _subprocess_apply(action, ctx, args, kwargs):
    # We with in the cache here to make sure archiver.load* puts things in the
    # right cache
    with ctx.cache:
        exe = action._bind(
            lambda: qiime2.sdk.Context(parent=ctx), {'type': 'asynchronous'})
        results = exe(*args, **kwargs)

        return results


def _run_parsl_action(action, ctx, execution_ctx, mapped_args, mapped_kwargs,
                      inputs=[]):
    """This is what the parsl app itself actually runs. It's basically just a
    wrapper around our QIIME 2 action. When this is initially called, args and
    kwargs may contain proxies that reference futures in inputs. By the time
    this starts executing, those futures will have resolved. We then need to
    take the resolved inputs and map the correct parts of them to the correct
    args/kwargs before calling the action with them.

    This is necessary because a single future in inputs will resolve into a
    Results object. We need to take singular Result objects off of that Results
    object and map them to the correct inputs for the action we want to call.
    """
    args = []
    for arg in mapped_args:
        unmapped = _unmap_arg(arg, inputs)
        args.append(unmapped)

    kwargs = {}
    for key, value in mapped_kwargs.items():
        unmapped = _unmap_arg(value, inputs)
        kwargs[key] = unmapped

    # We with in the cache here to make sure archiver.load* puts things in the
    # right cache
    with ctx.cache:
        exe = action._bind(
            lambda: qiime2.sdk.Context(parent=ctx), execution_ctx)
        results = exe(*args, **kwargs)

        # If we are running a pipeline, we need to create a future here because
        # the parsl join app the pipeline was running in is expected to return
        # a future, but we will have concrete results by this point if we are a
        # pipeline
        if isinstance(action, Pipeline) and ctx.parallel:
            return _create_future(results)

        return results


def _map_arg(arg, futures):
    """ Map a proxy artifact for input to a parsl action
    """

    # We add this future to the list and create a new proxy with its index as
    # its future.
    if isinstance(arg, Proxy):
        futures.append(arg._future_)
        mapped = arg.__class__(len(futures) - 1, arg._selector_)
    # We do the above but for all elements in the collection
    elif isinstance(arg, list) and _is_all_proxies(arg):
        mapped = []

        for proxy in arg:
            futures.append(proxy._future_)
            mapped.append(proxy.__class__(len(futures) - 1, proxy._selector_))
    elif isinstance(arg, dict) and _is_all_proxies(arg):
        mapped = {}

        for key, value in arg.items():
            futures.append(value._future_)
            mapped[key] = value.__class__(len(futures) - 1, value._selector_)
    # We just have a real artifact and don't need to map
    else:
        mapped = arg

    return mapped


def _unmap_arg(arg, inputs):
    """ Unmap a proxy artifact given to a parsl action
    """

    # We were hacky and set _future_ to be the index of this artifact in the
    # inputs list
    if isinstance(arg, Proxy):
        resolved_result = inputs[arg._future_]
        unmapped = arg._get_element_(resolved_result)
    # If we got a collection of proxies as the input we were even hackier and
    # added each proxy to the inputs list individually while having a list of
    # their indices in the args.
    elif isinstance(arg, list) and _is_all_proxies(arg):
        unmapped = []

        for proxy in arg:
            resolved_result = inputs[proxy._future_]
            unmapped.append(proxy._get_element_(resolved_result))
    elif isinstance(arg, dict) and _is_all_proxies(arg):
        unmapped = {}

        for key, value in arg.items():
            resolved_result = inputs[value._future_]
            unmapped[key] = value._get_element_(resolved_result)
    # We didn't have a proxy at all
    else:
        unmapped = arg

    return unmapped


def _is_all_proxies(collection):
    """ Returns whether the collection is all proxies or all artifacts.
        Raises a ValueError if there is a mix.
    """
    if isinstance(collection, dict):
        collection = list(collection.values())

    if all(isinstance(elem, Proxy) for elem in collection):
        return True

    if any(isinstance(elem, Proxy) for elem in collection):
        raise ValueError("Collection has mixed proxies and artifacts. "
                         "This is not allowed.")

    return False


@python_app
def _create_future(results):
    """ This is a bit of a dumb hack. It's just a way for us to make pipelines
    return a future which is what Parsl wants a join_app to return even though
    we will have real results at this point.
    """
    return results


class Action(metaclass=abc.ABCMeta):
    """QIIME 2 Action"""
    type = 'action'
    _ProvCaptureCls = archive.ActionProvenanceCapture

    __call__ = LateBindingAttribute('_dynamic_call')
    asynchronous = LateBindingAttribute('_dynamic_async')
    parallel = LateBindingAttribute('_dynamic_parsl')

    # Converts a callable's signature into its wrapper's signature (i.e.
    # converts the "view API" signature into the "artifact API" signature).
    # Accepts a callable as input and returns a callable as output with
    # converted signature.
    @abc.abstractmethod
    def _callable_sig_converter_(self, callable):
        raise NotImplementedError

    # Executes a callable on the provided `view_args`, wrapping and returning
    # the callable's outputs. In other words, executes the "view API", wrapping
    # and returning the outputs as the "artifact API". `view_args` is a dict
    # mapping parameter name to unwrapped value (i.e. view). `view_args`
    # contains an entry for each parameter accepted by the wrapper. It is the
    # executor's responsibility to perform any additional transformations on
    # these parameters, or provide extra parameters, in order to execute the
    # callable. `output_types` is an OrderedDict mapping output name to QIIME
    # type (e.g. semantic type).
    @abc.abstractmethod
    def _callable_executor_(self, scope, view_args, output_types):
        raise NotImplementedError

    # Private constructor
    @classmethod
    def _init(cls, callable, signature, plugin_id, name, description,
              citations, deprecated, examples):
        """

        Parameters
        ----------
        callable : callable
        signature : qiime2.core.type.Signature
        plugin_id : str
        name : str
            Human-readable name for this action.
        description : str
            Human-readable description for this action.

        """
        self = cls.__new__(cls)
        self.__init(callable, signature, plugin_id, name, description,
                    citations, deprecated, examples)
        return self

    # This "extra private" constructor is necessary because `Action` objects
    # can be initialized from a static (classmethod) context or on an
    # existing instance (see `_init` and `__setstate__`, respectively).
    def __init(self, callable, signature, plugin_id, name, description,
               citations, deprecated, examples):
        self._callable = callable
        self.signature = signature
        self.plugin_id = plugin_id
        self.name = name
        self.description = description
        self.citations = citations
        self.deprecated = deprecated
        self.examples = examples

        self.id = callable.__name__
        self._dynamic_call = self._get_callable_wrapper()
        self._dynamic_async = self._get_async_wrapper()
        # This a temp thing to play with parsl before integrating more deeply
        self._dynamic_parsl = self._get_parsl_wrapper()

    def __init__(self):
        raise NotImplementedError(
            "%s constructor is private." % self.__class__.__name__)

    @property
    def source(self):
        """
        The source code for the action's callable.

        Returns
        -------
        str
            The source code of this action's callable formatted as Markdown
            text.

        """
        try:
            source = inspect.getsource(self._callable)
        except OSError:
            raise TypeError(
                "Cannot retrieve source code for callable %r" %
                self._callable.__name__)
        return markdown_source_template % {'source': source}

    def get_import_path(self, include_self=True):
        path = f'qiime2.plugins.{self.plugin_id}.{self.type}s'
        if include_self:
            path += f'.{self.id}'
        return path

    def __repr__(self):
        return "<%s %s>" % (self.type, self.get_import_path())

    def __getstate__(self):
        return dill.dumps({
            'callable': self._callable,
            'signature': self.signature,
            'plugin_id': self.plugin_id,
            'name': self.name,
            'description': self.description,
            'citations': self.citations,
            'deprecated': self.deprecated,
            'examples': self.examples,
        })

    def __setstate__(self, state):
        self.__init(**dill.loads(state))

    def _bind(self, context_factory, execution_ctx={'type': 'synchronous'}):
        """Bind an action to a Context factory, returning a decorated function.

        This is a very primitive API and should be used primarily by the
        framework and very advanced interfaces which need deep control over
        the calling semantics of pipelines and garbage collection.

        The basic idea behind this is outlined as follows:

        Every action is defined as an *instance* that a plugin constructs.
        This means that `self` represents the internal details as to what
        the action is. If you need to associate additional state with the
        *application* of an action, you cannot mutate `self` without
        changing all future applications. So there needs to be an
        additional instance variable that can serve as the state of a given
        application. We call this a Context object. It is also important
        that each application of an action has *independent* state, so
        providing an instance of Context won't work. We need a factory.

        Parameterizing the context is necessary because it is possible for
        an action to call other actions. The details need to be coordinated
        behind the scenes to the user, so we can parameterize the behavior
        by providing different context factories to `bind` at different
        points in the "call stack".

        """
        def bound_callable(*args, **kwargs):
            # This function's signature is rewritten below using
            # `decorator.decorator`. When the signature is rewritten,
            # args[0] is the function whose signature was used to rewrite
            # this function's signature.
            args = args[1:]
            ctx = context_factory()
            # Set up a scope under which we can track destructable references
            # if something goes wrong, the __exit__ handler of this context
            # manager will clean up. (It also cleans up when things go right)
            with ctx as scope:
                provenance = self._ProvCaptureCls(
                    self.type, self.plugin_id, self.id, execution_ctx)
                scope.add_reference(provenance)

                if self.deprecated:
                    with qiime2.core.util.warning() as warn:
                        warn(self._build_deprecation_message(),
                             FutureWarning)

                # Type management
                collated_inputs = self.signature.collate_inputs(
                    *args, **kwargs)
                self.signature.check_types(**collated_inputs)
                output_types = self.signature.solve_output(**collated_inputs)
                callable_args = self.signature.coerce_user_input(
                    **collated_inputs)

                callable_args = \
                    self.signature.transform_and_add_callable_args_to_prov(
                        provenance, **callable_args)

                outputs = self._callable_executor_(
                    scope, callable_args, output_types, provenance)

                if len(outputs) != len(self.signature.outputs):
                    raise ValueError(
                        "Number of callable outputs must match number of "
                        "outputs defined in signature: %d != %d" %
                        (len(outputs), len(self.signature.outputs)))

                # Wrap in a Results object mapping output name to value so
                # users have access to outputs by name or position.
                results = qiime2.sdk.Results(
                    self.signature.outputs.keys(), outputs)

                return results

        bound_callable = self._rewrite_wrapper_signature(bound_callable)
        self._set_wrapper_properties(bound_callable)
        self._set_wrapper_name(bound_callable, self.id)
        return bound_callable

    def _get_callable_wrapper(self):
        # This is a "root" level invocation (not a nested call within a
        # pipeline), so no special factory is needed.
        callable_wrapper = self._bind(qiime2.sdk.Context)
        self._set_wrapper_name(callable_wrapper, '__call__')
        return callable_wrapper

    def _get_async_wrapper(self):
        def async_wrapper(*args, **kwargs):
            # TODO handle this better in the future, but stop the massive error
            # caused by MacOSX asynchronous runs for now.
            try:
                import matplotlib as plt
                if plt.rcParams['backend'].lower() == 'macosx':
                    raise EnvironmentError(backend_error_template %
                                           plt.matplotlib_fname())
            except ImportError:
                pass

            # This function's signature is rewritten below using
            # `decorator.decorator`. When the signature is rewritten, args[0]
            # is the function whose signature was used to rewrite this
            # function's signature.
            args = args[1:]

            pool = concurrent.futures.ProcessPoolExecutor(max_workers=1)
            future = pool.submit(
                _subprocess_apply, self, qiime2.sdk.Context(), args, kwargs)
            # TODO: pool.shutdown(wait=False) caused the child process to
            # hang unrecoverably. This seems to be a bug in Python 3.7
            # It's probably best to gut concurrent.futures entirely, so we're
            # ignoring the resource leakage for the moment.
            return future

        async_wrapper = self._rewrite_wrapper_signature(async_wrapper)
        self._set_wrapper_properties(async_wrapper)
        self._set_wrapper_name(async_wrapper, 'asynchronous')
        return async_wrapper

    def _bind_parsl(self, ctx, *args, **kwargs):
        futures = []
        mapped_args = []
        mapped_kwargs = {}

        # If this is the first time we called _bind_parsl on a pipeline, the
        # first argument will be the callable for the pipeline which we do not
        # want to pass on in this manner, so we skip it.
        if len(args) >= 1 and callable(args[0]):
            args = args[1:]

        # Parsl will queue up apps with futures as their arguments then not
        # execute the apps until the futures are resolved. This is an extremely
        # handy feature, but QIIME 2 does not play nice with it out of the box.
        # You can look in qiime2/sdk/proxy.py for some more details on how this
        # is working, but we are basically taking future QIIME 2 results and
        # mapping them to the correct inputs in the action we are trying to
        # call. This is necessary if we are running a pipeline in particular
        # because the inputs to the next action could contain outputs from the
        # last action that might not be resolved yet because Parsl may be
        # queueing the next action before the last one has completed.
        for arg in args:
            mapped = _map_arg(arg, futures)
            mapped_args.append(mapped)

        for key, value in kwargs.items():
            mapped = _map_arg(value, futures)
            mapped_kwargs[key] = mapped

        # If the user specified a particular executor for a this action
        # determine that here
        executor = ctx.action_executor_mapping.get(self.id, 'default')
        execution_ctx = {'type': 'parsl'}

        # Pipelines run in join apps and are a sort of synchronization point
        # right now. Unfortunately it is not currently possible to make say a
        # pipeline that calls two other pipelines within it and execute both of
        # those internal pipelines simultaneously.
        if isinstance(self, qiime2.sdk.action.Pipeline):
            # If ctx._parent is None then this is the root pipeline and we want
            # to dispatch it to a join_app
            execution_ctx['parsl_type'] = 'DFK'
            if ctx._parent is None:
                # NOTE: Do not make this a python_app(join=True). We need it to
                # run in the parsl main thread
                future = join_app()(
                        _run_parsl_action)(self, ctx, execution_ctx,
                                           mapped_args, mapped_kwargs,
                                           inputs=futures)
            # If there is a parent then this is not the root pipeline and we
            # want to just _bind it with a parallel context. The fact that
            # parallel is set on the context will cause ctx.get_action calls in
            # the pipeline to use the action's _bind_parsl method.
            else:
                return self._bind(lambda: qiime2.sdk.Context(ctx),
                                  execution_ctx=execution_ctx)(*args, **kwargs)
        else:
            execution_ctx['parsl_type'] = \
                ctx.executor_name_type_mapping[executor]
            future = python_app(
                executors=[executor])(
                    _run_parsl_action)(self, ctx, execution_ctx,
                                       mapped_args, mapped_kwargs,
                                       inputs=futures)

        collated_input = self.signature.collate_inputs(*args, **kwargs)
        output_types = self.signature.solve_output(**collated_input)

        # Again, we return a set of futures not a set of real results
        return qiime2.sdk.proxy.ProxyResults(future, output_types)

    def _get_parsl_wrapper(self):
        def parsl_wrapper(*args, **kwargs):
            # TODO: Maybe make this a warning instead?
            if not isinstance(self, Pipeline):
                raise ValueError('Only pipelines may be run in parallel')

            return self._bind_parsl(qiime2.sdk.Context(parallel=True), *args,
                                    **kwargs)

        parsl_wrapper = self._rewrite_wrapper_signature(parsl_wrapper)
        self._set_wrapper_properties(parsl_wrapper)
        self._set_wrapper_name(parsl_wrapper, 'parsl')
        return parsl_wrapper

    def _rewrite_wrapper_signature(self, wrapper):
        # Convert the callable's signature into the wrapper's signature and set
        # it on the wrapper.
        return decorator.decorator(
            wrapper, self._callable_sig_converter_(self._callable))

    def _set_wrapper_name(self, wrapper, name):
        wrapper.__name__ = wrapper.__qualname__ = name

    def _set_wrapper_properties(self, wrapper):
        wrapper.__module__ = self.get_import_path(include_self=False)
        wrapper.__doc__ = self._build_numpydoc()
        wrapper.__annotations__ = self._build_annotations()
        # This is necessary so that `inspect` doesn't display the wrapped
        # function's annotations (the annotations apply to the "view API" and
        # not the "artifact API").
        del wrapper.__wrapped__

    def _build_annotations(self):
        annotations = {}
        for name, spec in self.signature.signature_order.items():
            annotations[name] = spec.qiime_type

        output = []
        for spec in self.signature.outputs.values():
            output.append(spec.qiime_type)
        output = tuple(output)

        annotations["return"] = output

        return annotations

    def _build_numpydoc(self):
        numpydoc = []
        numpydoc.append(textwrap.fill(self.name, width=75))
        if self.deprecated:
            base_msg = textwrap.indent(
                textwrap.fill(self._build_deprecation_message(), width=72),
                '   ')
            numpydoc.append('.. deprecated::\n' + base_msg)
        numpydoc.append(textwrap.fill(self.description, width=75))

        sig = self.signature
        parameters = self._build_section("Parameters", sig.signature_order)
        returns = self._build_section("Returns", sig.outputs)

        # TODO: include Usage-rendered examples here

        for section in (parameters, returns):
            if section:
                numpydoc.append(section)

        return '\n\n'.join(numpydoc) + '\n'

    def _build_section(self, header, iterable):
        section = []

        if iterable:
            section.append(header)
            section.append('-'*len(header))
            for key, value in iterable.items():
                variable_line = (
                    "{item} : {type}".format(item=key, type=value.qiime_type))
                if value.has_default():
                    variable_line += ", optional"
                section.append(variable_line)
                if value.has_description():
                    section.append(textwrap.indent(textwrap.fill(
                        str(value.description), width=71), '    '))

        return '\n'.join(section).strip()

    def _build_deprecation_message(self):
        return (f'This {self.type.title()} is deprecated and will be removed '
                'in a future version of this plugin.')


class Method(Action):
    """QIIME 2 Method"""

    type = 'method'

    # Abstract method implementations:

    def _callable_sig_converter_(self, callable):
        # No conversion necessary.
        return callable

    def _callable_executor_(self, scope, view_args, output_types, provenance):
        output_views = self._callable(**view_args)
        output_views = tuplize(output_views)

        # TODO this won't work if the user has annotated their "view API" to
        # return a `typing.Tuple` with some number of components. Python will
        # return a tuple when there are multiple return values, and this length
        # check will fail because the tuple as a whole should be matched up to
        # a single output type instead of its components. This is an edgecase
        # due to how Python handles multiple returns, and can be worked around
        # by using something like `typing.List` instead.
        if len(output_views) != len(output_types):
            raise TypeError(
                "Number of output views must match number of output "
                "semantic types: %d != %d"
                % (len(output_views), len(output_types)))

        output_artifacts = \
            self.signature.coerce_given_outputs(output_views, output_types,
                                                scope, provenance)

        return tuple(output_artifacts)

    @classmethod
    def _init(cls, callable, inputs, parameters, outputs, plugin_id, name,
              description, input_descriptions, parameter_descriptions,
              output_descriptions, citations, deprecated, examples):
        signature = qtype.MethodSignature(callable, inputs, parameters,
                                          outputs, input_descriptions,
                                          parameter_descriptions,
                                          output_descriptions)
        return super()._init(callable, signature, plugin_id, name, description,
                             citations, deprecated, examples)


class Visualizer(Action):
    """QIIME 2 Visualizer"""

    type = 'visualizer'

    # Abstract method implementations:

    def _callable_sig_converter_(self, callable):
        return DropFirstParameter.from_function(callable)

    def _callable_executor_(self, scope, view_args, output_types, provenance):
        # TODO use qiime2.plugin.OutPath when it exists, and update visualizers
        # to work with OutPath instead of str. Visualization._from_data_dir
        # will also need to be updated to support OutPath instead of str.
        with tempfile.TemporaryDirectory(prefix='qiime2-temp-') as temp_dir:
            ret_val = self._callable(output_dir=temp_dir, **view_args)
            if ret_val is not None:
                raise TypeError(
                    "Visualizer %r should not return anything. "
                    "Received %r as a return value." % (self, ret_val))
            provenance.output_name = 'visualization'
            viz = qiime2.sdk.Visualization._from_data_dir(temp_dir,
                                                          provenance)
            viz = scope.add_parent_reference(viz)

            return (viz, )

    @classmethod
    def _init(cls, callable, inputs, parameters, plugin_id, name, description,
              input_descriptions, parameter_descriptions, citations,
              deprecated, examples):
        signature = qtype.VisualizerSignature(callable, inputs, parameters,
                                              input_descriptions,
                                              parameter_descriptions)
        return super()._init(callable, signature, plugin_id, name, description,
                             citations, deprecated, examples)


class Pipeline(Action):
    """QIIME 2 Pipeline"""
    type = 'pipeline'
    _ProvCaptureCls = archive.PipelineProvenanceCapture

    def _callable_sig_converter_(self, callable):
        return DropFirstParameter.from_function(callable)

    def _callable_executor_(self, scope, view_args, output_types, provenance):
        outputs = self._callable(scope.ctx, **view_args)
        # Just make sure we have an iterable even if there was only one output
        outputs = tuplize(outputs)
        # Make sure any collections returned are in the form of
        # ResultCollections and that futures are resolved
        #
        # TODO: Ideally we would not need to resolve futures here as this
        # prevents us from properly parallelizing nested pipelines
        outputs = self._coerce_pipeline_outputs(outputs)

        for output in outputs:
            if isinstance(output, qiime2.sdk.ResultCollection):
                for elem in output.values():
                    if not isinstance(elem, qiime2.sdk.Result):
                        raise TypeError("Pipelines must return `Result` "
                                        "objects, not %s" % (type(elem), ))
            elif not isinstance(output, qiime2.sdk.Result):
                raise TypeError("Pipelines must return `Result` objects, "
                                "not %s" % (type(output), ))

        # This condition *is* tested by the caller of _callable_executor_, but
        # the kinds of errors a plugin developer see will make more sense if
        # this check happens before the subtype check. Otherwise forgetting an
        # output would more likely error as a wrong type, which while correct,
        # isn't root of the problem.
        if len(outputs) != len(output_types):
            raise TypeError(
                "Number of outputs must match number of output "
                "semantic types: %d != %d"
                % (len(outputs), len(output_types)))

        results = []
        for output, (name, spec) in zip(outputs, output_types.items()):
            # If we don't have a Result, we should have a collection, if we
            # have neither, or our types just don't match up, something bad
            # happened
            if isinstance(output, qiime2.sdk.Result) and \
                    (output.type <= spec.qiime_type):
                prov = provenance.fork(name, output)
                scope.add_reference(prov)

                aliased_result = output._alias(prov)
                aliased_result = scope.add_parent_reference(aliased_result)

                results.append(aliased_result)
            elif spec.qiime_type.name == 'Collection' and \
                    output.collection in spec.qiime_type:
                size = len(output)
                aliased_output = qiime2.sdk.ResultCollection()
                for idx, (key, value) in enumerate(output.items()):
                    collection_name = create_collection_name(
                        name=name, key=key, idx=idx, size=size)
                    prov = provenance.fork(collection_name, value)
                    scope.add_reference(prov)

                    aliased_result = value._alias(prov)
                    aliased_result = scope.add_parent_reference(aliased_result)
                    aliased_output[str(key)] = aliased_result

                results.append(aliased_output)
            else:
                _type = output.type if isinstance(output, qiime2.sdk.Result) \
                    else type(output)
                raise TypeError(
                    "Expected output type %r, received %r" %
                    (spec.qiime_type, _type))

        if len(results) != len(self.signature.outputs):
            raise ValueError(
                "Number of callable outputs must match number of "
                "outputs defined in signature: %d != %d" %
                (len(results), len(self.signature.outputs)))

        return tuple(results)

    def _coerce_pipeline_outputs(self, outputs):
        """Ensure all futures are resolved and all collections are of type
           ResultCollection
        """
        coerced_outputs = []

        for output in outputs:
            # Handle proxy outputs
            if isinstance(output, Proxy):
                output = output.result()

            # Handle collection outputs
            if isinstance(output, dict) or \
                    isinstance(output, list):
                output = qiime2.sdk.ResultCollection(output)

                # Handle proxies as elements of collections
                for key, value in output.items():
                    if isinstance(value, Proxy):
                        output[key] = value.result()

            coerced_outputs.append(output)

        return tuple(coerced_outputs)

    @classmethod
    def _init(cls, callable, inputs, parameters, outputs, plugin_id, name,
              description, input_descriptions, parameter_descriptions,
              output_descriptions, citations, deprecated, examples):
        signature = qtype.PipelineSignature(callable, inputs, parameters,
                                            outputs, input_descriptions,
                                            parameter_descriptions,
                                            output_descriptions)
        return super()._init(callable, signature, plugin_id, name, description,
                             citations, deprecated, examples)


markdown_source_template = """
```python
%(source)s
```
"""

# TODO add unit test for callables raising this
backend_error_template = """
Your current matplotlib backend (MacOSX) does not work with asynchronous calls.
A recommended backend is Agg, and can be changed by modifying your
matplotlibrc "backend" parameter, which can be found at: \n\n %s
"""



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import os
import tempfile
import unittest

import pandas as pd

import qiime2
from qiime2.core.cache import Cache
from qiime2.core.testing.type import IntSequence1, SingleInt
from qiime2.core.testing.util import get_dummy_plugin, PipelineError
from qiime2.sdk.result import Artifact
from qiime2.sdk.parallel_config import ParallelConfig
from qiime2.core.util import load_action_yaml


def _load_alias_uuid(result):
    return load_action_yaml(result._archiver.path)['action']['alias-of']


def _load_nested_alias_uuid(result, cache):
    alias_uuid = _load_alias_uuid(result)
    aliased_result = qiime2.sdk.Result.load(
        os.path.join(cache.data, alias_uuid))
    return _load_alias_uuid(aliased_result)


def _load_alias_uuids(collection):
    uuids = []

    for result in collection.values():
        uuids.append(_load_alias_uuid(result))

    return uuids


def _load_nested_alias_uuids(collection, cache):
    alias_uuids = _load_alias_uuids(collection)

    # load_alias_uuids is expecting a dictionary, so just make a dictionary
    # here so it gets what it wants
    alias_results = {}
    for idx, alias_uuid in enumerate(alias_uuids):
        alias_results[idx] = \
            qiime2.sdk.Result.load(os.path.join(cache.data, alias_uuid))

    return _load_alias_uuids(alias_results)


class TestPipelineResumption(unittest.TestCase):
    def setUp(self):
        # Get our pipeline
        self.plugin = get_dummy_plugin()
        self.pipeline = self.plugin.pipelines['resumable_varied_pipeline']
        self.nested_pipeline = \
            self.plugin.pipelines['resumable_nested_varied_pipeline']

        # Create temp test dir
        self.test_dir = tempfile.TemporaryDirectory(prefix='qiime2-test-temp-')

        # Create cache and pool
        self.cache = Cache(os.path.join(self.test_dir.name, 'cache'))
        self.pool = self.cache.create_pool('pool')

        # Create artifacts
        self.ints1 = {'1': Artifact.import_data(SingleInt, 0),
                      '2': Artifact.import_data(SingleInt, 1)}
        self.ints1_2 = {'3': Artifact.import_data(SingleInt, 1),
                        '4': Artifact.import_data(SingleInt, 2)}
        self.ints2 = [Artifact.import_data(IntSequence1, [0, 1, 2]),
                      Artifact.import_data(IntSequence1, [3, 4, 5])]
        self.int1 = Artifact.import_data(SingleInt, 42)
        self.int2 = Artifact.import_data(SingleInt, 43)

        # Create metadata
        df1 = pd.DataFrame({'a': ['1', '2', '3']},
                           index=pd.Index(['0', '1', '2'], name='feature ID'))
        self.md1 = qiime2.Metadata(df1)
        df2 = pd.DataFrame({'b': ['4', '5', '6']},
                           index=pd.Index(['0', '1', '2'], name='feature ID'))
        self.md2 = qiime2.Metadata(df2)

    def tearDown(self):
        """Remove our cache and all that from last test
        """
        self.test_dir.cleanup()

    def test_resumable_pipeline_no_pool(self):
        with self.cache:
            with self.assertRaises(PipelineError) as e:
                self.pipeline(
                    self.ints1, self.ints2, self.md1, self.int1, 'Hi',
                    fail=True)

            ints1_uuids, ints2_uuids, int1_uuid, list_uuids, dict_uuids, \
                identity_uuid, viz_uuid = e.exception.uuids

            ints1_ret, ints2_ret, int1_ret, list_ret, dict_ret, identity_ret, \
                viz_ret = self.pipeline(
                    self.ints1, self.ints2, self.md1, self.int1, 'Hi')

            complete_ints1_uuids = _load_alias_uuids(ints1_ret)
            complete_ints2_uuids = _load_alias_uuids(ints2_ret)
            complete_int1_uuid = _load_alias_uuid(int1_ret)
            complete_list_uuids = _load_alias_uuids(list_ret)
            complete_dict_uuids = _load_alias_uuids(dict_ret)
            complete_identity_uuid = _load_alias_uuid(identity_ret)
            complete_viz_uuid = _load_alias_uuid(viz_ret)

            # Nothing should have been recycled because we didn't use a pool
            self.assertNotEqual(ints1_uuids, complete_ints1_uuids)
            self.assertNotEqual(ints2_uuids, complete_ints2_uuids)
            self.assertNotEqual(int1_uuid, complete_int1_uuid)
            self.assertNotEqual(list_uuids, complete_list_uuids)
            self.assertNotEqual(dict_uuids, complete_dict_uuids)
            self.assertNotEqual(identity_uuid, complete_identity_uuid)
            self.assertNotEqual(viz_uuid, complete_viz_uuid)

    def test_resumable_pipeline(self):
        with self.pool:
            with self.assertRaises(PipelineError) as e:
                self.pipeline(
                    self.ints1, self.ints2, self.md1, self.int1, 'Hi',
                    fail=True)

            ints1_uuids, ints2_uuids, int1_uuid, list_uuids, dict_uuids, \
                identity_uuid, viz_uuid = e.exception.uuids

            ints1_ret, ints2_ret, int1_ret, list_ret, dict_ret, \
                identity_ret, viz_ret = self.pipeline(
                    self.ints1, self.ints2, self.md1, self.int1, 'Hi')

            complete_ints1_uuids = _load_alias_uuids(ints1_ret)
            complete_ints2_uuids = _load_alias_uuids(ints2_ret)
            complete_int1_uuid = _load_alias_uuid(int1_ret)
            complete_list_uuids = _load_alias_uuids(list_ret)
            complete_dict_uuids = _load_alias_uuids(dict_ret)
            complete_identity_uuid = _load_alias_uuid(identity_ret)
            complete_viz_uuid = _load_alias_uuid(viz_ret)

            # Assert that the artifacts returned by the completed run of
            # the pipeline are aliases of the artifacts created by the
            # first failed run
            self.assertEqual(ints1_uuids, complete_ints1_uuids)
            self.assertEqual(ints2_uuids, complete_ints2_uuids)
            self.assertEqual(int1_uuid, complete_int1_uuid)
            self.assertEqual(list_uuids, complete_list_uuids)
            self.assertEqual(dict_uuids, complete_dict_uuids)
            self.assertEqual(identity_uuid, complete_identity_uuid)
            self.assertEqual(viz_uuid, complete_viz_uuid)

    def test_resumable_pipeline_parallel(self):
        with self.pool:
            with self.assertRaises(PipelineError) as e:
                with ParallelConfig():
                    future = self.pipeline.parallel(
                        self.ints1, self.ints2, self.md1, self.int1, 'Hi',
                        fail=True)
                    future._result()

            ints1_uuids, ints2_uuids, int1_uuid, list_uuids, dict_uuids, \
                identity_uuid, viz_uuid = e.exception.uuids

            with ParallelConfig():
                future = self.pipeline.parallel(
                    self.ints1, self.ints2, self.md1, self.int1, 'Hi')
                ints1_ret, ints2_ret, int1_ret, list_ret, dict_ret, \
                    identity_ret, viz_ret = future._result()

            complete_ints1_uuids = _load_alias_uuids(ints1_ret)
            complete_ints2_uuids = _load_alias_uuids(ints2_ret)
            complete_int1_uuid = _load_alias_uuid(int1_ret)
            complete_list_uuids = _load_alias_uuids(list_ret)
            complete_dict_uuids = _load_alias_uuids(dict_ret)
            complete_identity_uuid = _load_alias_uuid(identity_ret)
            complete_viz_uuid = _load_alias_uuid(viz_ret)

            # Assert that the artifacts returned by the completed run of
            # the pipeline are aliases of the artifacts created by the
            # first failed run
            self.assertEqual(ints1_uuids, complete_ints1_uuids)
            self.assertEqual(ints2_uuids, complete_ints2_uuids)
            self.assertEqual(int1_uuid, complete_int1_uuid)
            self.assertEqual(list_uuids, complete_list_uuids)
            self.assertEqual(dict_uuids, complete_dict_uuids)
            self.assertEqual(identity_uuid, complete_identity_uuid)
            self.assertEqual(viz_uuid, complete_viz_uuid)

    def test_resumable_pipeline_artifact_varies(self):
        with self.pool:
            with self.assertRaises(PipelineError) as e:
                self.pipeline(
                    self.ints1, self.ints2, self.md1, self.int1, 'Hi',
                    fail=True)

            ints1_uuids, ints2_uuids, int1_uuid, list_uuids, dict_uuids, \
                identity_uuid, viz_uuid = e.exception.uuids

            # Pass int2 instead of int1
            ints1_ret, ints2_ret, int1_ret, list_ret, dict_ret, \
                identity_ret, viz_ret = self.pipeline(
                    self.ints1, self.ints2, self.md1, self.int2, 'Hi')

            complete_ints1_uuids = _load_alias_uuids(ints1_ret)
            complete_ints2_uuids = _load_alias_uuids(ints2_ret)
            complete_int1_uuid = _load_alias_uuid(int1_ret)
            complete_list_uuids = _load_alias_uuids(list_ret)
            complete_dict_uuids = _load_alias_uuids(dict_ret)
            complete_identity_uuid = _load_alias_uuid(identity_ret)
            complete_viz_uuid = _load_alias_uuid(viz_ret)

            # Assert that the artifacts returned by the completed pipeline that
            # are implicated by the changed input are not aliases while the
            # others are
            self.assertNotEqual(ints1_uuids, complete_ints1_uuids)
            self.assertNotEqual(ints2_uuids, complete_ints2_uuids)
            self.assertNotEqual(int1_uuid, complete_int1_uuid)
            self.assertNotEqual(list_uuids, complete_list_uuids)
            self.assertEqual(dict_uuids, complete_dict_uuids)
            self.assertEqual(identity_uuid, complete_identity_uuid)
            self.assertEqual(viz_uuid, complete_viz_uuid)

    def test_resumable_pipeline_artifact_varies_parallel(self):
        with self.pool:
            with self.assertRaises(PipelineError) as e:
                with ParallelConfig():
                    future = self.pipeline.parallel(
                        self.ints1, self.ints2, self.md1, self.int1, 'Hi',
                        fail=True)
                    future._result()

            ints1_uuids, ints2_uuids, int1_uuid, list_uuids, dict_uuids, \
                identity_uuid, viz_uuid = e.exception.uuids

            # Pass int2 instead of int1
            with ParallelConfig():
                future = self.pipeline.parallel(
                    self.ints1, self.ints2, self.md1, self.int2, 'Hi')
                ints1_ret, ints2_ret, int1_ret, list_ret, dict_ret, \
                    identity_ret, viz_ret = future._result()

            complete_ints1_uuids = _load_alias_uuids(ints1_ret)
            complete_ints2_uuids = _load_alias_uuids(ints2_ret)
            complete_int1_uuid = _load_alias_uuid(int1_ret)
            complete_list_uuids = _load_alias_uuids(list_ret)
            complete_dict_uuids = _load_alias_uuids(dict_ret)
            complete_identity_uuid = _load_alias_uuid(identity_ret)
            complete_viz_uuid = _load_alias_uuid(viz_ret)

            # Assert that the artifacts returned by the completed pipeline that
            # are implicated by the changed input are not aliases while the
            # others are
            self.assertNotEqual(ints1_uuids, complete_ints1_uuids)
            self.assertNotEqual(ints2_uuids, complete_ints2_uuids)
            self.assertNotEqual(int1_uuid, complete_int1_uuid)
            self.assertNotEqual(list_uuids, complete_list_uuids)
            self.assertEqual(dict_uuids, complete_dict_uuids)
            self.assertEqual(identity_uuid, complete_identity_uuid)
            self.assertEqual(viz_uuid, complete_viz_uuid)

    def test_resumable_pipeline_collection_varies(self):
        with self.pool:
            with self.assertRaises(PipelineError) as e:
                self.pipeline(
                    self.ints1, self.ints2, self.md1, self.int1, 'Hi',
                    fail=True)

            ints1_uuids, ints2_uuids, int1_uuid, list_uuids, dict_uuids, \
                identity_uuid, viz_uuid = e.exception.uuids

            # Pass ints1_2 instead of ints1
            ints1_ret, ints2_ret, int1_ret, list_ret, dict_ret, \
                identity_ret, viz_ret = self.pipeline(
                    self.ints1_2, self.ints2, self.md1, self.int1, 'Hi')

            complete_ints1_uuids = _load_alias_uuids(ints1_ret)
            complete_ints2_uuids = _load_alias_uuids(ints2_ret)
            complete_int1_uuid = _load_alias_uuid(int1_ret)
            complete_list_uuids = _load_alias_uuids(list_ret)
            complete_dict_uuids = _load_alias_uuids(dict_ret)
            complete_identity_uuid = _load_alias_uuid(identity_ret)
            complete_viz_uuid = _load_alias_uuid(viz_ret)

            # Assert that the artifacts returned by the completed pipeline that
            # are implicated by the changed input are not aliases while the
            # others are
            self.assertNotEqual(ints1_uuids, complete_ints1_uuids)
            self.assertNotEqual(ints2_uuids, complete_ints2_uuids)
            self.assertNotEqual(int1_uuid, complete_int1_uuid)
            self.assertNotEqual(list_uuids, complete_list_uuids)
            self.assertNotEqual(dict_uuids, complete_dict_uuids)
            self.assertEqual(identity_uuid, complete_identity_uuid)
            self.assertEqual(viz_uuid, complete_viz_uuid)

    def test_resumable_pipeline_collection_varies_parallel(self):
        with self.pool:
            with self.assertRaises(PipelineError) as e:
                with ParallelConfig():
                    future = self.pipeline.parallel(
                        self.ints1, self.ints2, self.md1, self.int1, 'Hi',
                        fail=True)
                    future._result()

            ints1_uuids, ints2_uuids, int1_uuid, list_uuids, dict_uuids, \
                identity_uuid, viz_uuid = e.exception.uuids

            # Pass ints1_2 instead of ints1
            with ParallelConfig():
                future = self.pipeline.parallel(
                    self.ints1_2, self.ints2, self.md1, self.int2, 'Hi')
                ints1_ret, ints2_ret, int1_ret, list_ret, dict_ret, \
                    identity_ret, viz_ret = future._result()

            complete_ints1_uuids = _load_alias_uuids(ints1_ret)
            complete_ints2_uuids = _load_alias_uuids(ints2_ret)
            complete_int1_uuid = _load_alias_uuid(int1_ret)
            complete_list_uuids = _load_alias_uuids(list_ret)
            complete_dict_uuids = _load_alias_uuids(dict_ret)
            complete_identity_uuid = _load_alias_uuid(identity_ret)
            complete_viz_uuid = _load_alias_uuid(viz_ret)

            # Assert that the artifacts returned by the completed pipeline that
            # are implicated by the changed input are not aliases while the
            # others are
            self.assertNotEqual(ints1_uuids, complete_ints1_uuids)
            self.assertNotEqual(ints2_uuids, complete_ints2_uuids)
            self.assertNotEqual(int1_uuid, complete_int1_uuid)
            self.assertNotEqual(list_uuids, complete_list_uuids)
            self.assertNotEqual(dict_uuids, complete_dict_uuids)
            self.assertEqual(identity_uuid, complete_identity_uuid)
            self.assertEqual(viz_uuid, complete_viz_uuid)

    def test_resumable_pipeline_str_varies(self):
        with self.pool:
            with self.assertRaises(PipelineError) as e:
                self.pipeline(
                    self.ints1, self.ints2, self.md1, self.int1, 'Hi',
                    fail=True)

            ints1_uuids, ints2_uuids, int1_uuid, list_uuids, dict_uuids, \
                identity_uuid, viz_uuid = e.exception.uuids

            # Pass in Bye instead of Hi
            ints1_ret, ints2_ret, int1_ret, list_ret, dict_ret, \
                identity_ret, viz_ret = self.pipeline(
                    self.ints1, self.ints2, self.md1, self.int1, 'Bye')

            complete_ints1_uuids = _load_alias_uuids(ints1_ret)
            complete_ints2_uuids = _load_alias_uuids(ints2_ret)
            complete_int1_uuid = _load_alias_uuid(int1_ret)
            complete_list_uuids = _load_alias_uuids(list_ret)
            complete_dict_uuids = _load_alias_uuids(dict_ret)
            complete_identity_uuid = _load_alias_uuid(identity_ret)
            complete_viz_uuid = _load_alias_uuid(viz_ret)

            # Assert that the artifacts returned by the completed pipeline that
            # are implicated by the changed input are not aliases while the
            # others are
            self.assertNotEqual(ints1_uuids, complete_ints1_uuids)
            self.assertNotEqual(ints2_uuids, complete_ints2_uuids)
            self.assertNotEqual(int1_uuid, complete_int1_uuid)
            self.assertNotEqual(list_uuids, complete_list_uuids)
            self.assertEqual(dict_uuids, complete_dict_uuids)
            self.assertEqual(identity_uuid, complete_identity_uuid)
            self.assertEqual(viz_uuid, complete_viz_uuid)

    def test_resumable_pipeline_str_varies_parallel(self):
        with self.pool:
            with self.assertRaises(PipelineError) as e:
                with ParallelConfig():
                    future = self.pipeline.parallel(
                        self.ints1, self.ints2, self.md1, self.int1, 'Hi',
                        fail=True)
                    future._result()

            ints1_uuids, ints2_uuids, int1_uuid, list_uuids, dict_uuids, \
                identity_uuid, viz_uuid = e.exception.uuids

            # Pass in Bye instead of Hi
            with ParallelConfig():
                future = self.pipeline.parallel(
                    self.ints1, self.ints2, self.md1, self.int1, 'Bye')
                ints1_ret, ints2_ret, int1_ret, list_ret, dict_ret, \
                    identity_ret, viz_ret = future._result()

            complete_ints1_uuids = _load_alias_uuids(ints1_ret)
            complete_ints2_uuids = _load_alias_uuids(ints2_ret)
            complete_int1_uuid = _load_alias_uuid(int1_ret)
            complete_list_uuids = _load_alias_uuids(list_ret)
            complete_dict_uuids = _load_alias_uuids(dict_ret)
            complete_identity_uuid = _load_alias_uuid(identity_ret)
            complete_viz_uuid = _load_alias_uuid(viz_ret)

            # Assert that the artifacts returned by the completed pipeline that
            # are implicated by the changed input are not aliases while the
            # others are
            self.assertNotEqual(ints1_uuids, complete_ints1_uuids)
            self.assertNotEqual(ints2_uuids, complete_ints2_uuids)
            self.assertNotEqual(int1_uuid, complete_int1_uuid)
            self.assertNotEqual(list_uuids, complete_list_uuids)
            self.assertEqual(dict_uuids, complete_dict_uuids)
            self.assertEqual(identity_uuid, complete_identity_uuid)
            self.assertEqual(viz_uuid, complete_viz_uuid)

    def test_resumable_pipeline_md_varies(self):
        with self.pool:
            with self.assertRaises(PipelineError) as e:
                self.pipeline(
                    self.ints1, self.ints2, self.md1, self.int1, 'Hi',
                    fail=True)

            ints1_uuids, ints2_uuids, int1_uuid, list_uuids, dict_uuids, \
                identity_uuid, viz_uuid = e.exception.uuids

            # Pass in md2 instead of md1
            ints1_ret, ints2_ret, int1_ret, list_ret, dict_ret, \
                identity_ret, viz_ret = self.pipeline(
                    self.ints1, self.ints2, self.md2, self.int1, 'Hi')

            complete_ints1_uuids = _load_alias_uuids(ints1_ret)
            complete_ints2_uuids = _load_alias_uuids(ints2_ret)
            complete_int1_uuid = _load_alias_uuid(int1_ret)
            complete_list_uuids = _load_alias_uuids(list_ret)
            complete_dict_uuids = _load_alias_uuids(dict_ret)
            complete_identity_uuid = _load_alias_uuid(identity_ret)
            complete_viz_uuid = _load_alias_uuid(viz_ret)

            # Assert that the artifacts returned by the completed pipeline that
            # are implicated by the changed input are not aliases while the
            # others are
            self.assertEqual(ints1_uuids, complete_ints1_uuids)
            self.assertEqual(ints2_uuids, complete_ints2_uuids)
            self.assertEqual(int1_uuid, complete_int1_uuid)
            self.assertEqual(list_uuids, complete_list_uuids)
            self.assertEqual(dict_uuids, complete_dict_uuids)
            self.assertNotEqual(identity_uuid, complete_identity_uuid)
            self.assertEqual(viz_uuid, complete_viz_uuid)

    def test_resumable_pipeline_md_varies_parallel(self):
        with self.pool:
            with self.assertRaises(PipelineError) as e:
                with ParallelConfig():
                    future = self.pipeline.parallel(
                        self.ints1, self.ints2, self.md1, self.int1, 'Hi',
                        fail=True)
                    future._result()

            ints1_uuids, ints2_uuids, int1_uuid, list_uuids, dict_uuids, \
                identity_uuid, viz_uuid = e.exception.uuids

            # Pass in md2 instead of md1
            with ParallelConfig():
                future = self.pipeline.parallel(
                    self.ints1, self.ints2, self.md2, self.int1, 'Hi')
                ints1_ret, ints2_ret, int1_ret, list_ret, dict_ret, \
                    identity_ret, viz_ret = future._result()

            complete_ints1_uuids = _load_alias_uuids(ints1_ret)
            complete_ints2_uuids = _load_alias_uuids(ints2_ret)
            complete_int1_uuid = _load_alias_uuid(int1_ret)
            complete_list_uuids = _load_alias_uuids(list_ret)
            complete_dict_uuids = _load_alias_uuids(dict_ret)
            complete_identity_uuid = _load_alias_uuid(identity_ret)
            complete_viz_uuid = _load_alias_uuid(viz_ret)

            # Assert that the artifacts returned by the completed pipeline that
            # are implicated by the changed input are not aliases while the
            # others are
            self.assertEqual(ints1_uuids, complete_ints1_uuids)
            self.assertEqual(ints2_uuids, complete_ints2_uuids)
            self.assertEqual(int1_uuid, complete_int1_uuid)
            self.assertEqual(list_uuids, complete_list_uuids)
            self.assertEqual(dict_uuids, complete_dict_uuids)
            self.assertNotEqual(identity_uuid, complete_identity_uuid)
            self.assertEqual(viz_uuid, complete_viz_uuid)

    def test_nested_resumable_pipeline(self):
        with self.pool:
            with self.assertRaises(PipelineError) as e:
                self.nested_pipeline(
                    self.ints1, self.ints2, self.md1, self.int1, 'Hi',
                    fail=True)

            ints1_uuids, ints2_uuids, int1_uuid, list_uuids, dict_uuids, \
                identity_uuid, viz_uuid = e.exception.uuids

            # We now run the not nested version. This will be able to reuse the
            # returns from varied_method
            ints1_ret, ints2_ret, int1_ret, list_ret, dict_ret, \
                identity_ret, viz_ret = self.nested_pipeline(
                    self.ints1, self.ints2, self.md1, self.int1, 'Hi')

            complete_ints1_uuids = _load_nested_alias_uuids(
                ints1_ret, self.cache)
            complete_ints2_uuids = _load_nested_alias_uuids(
                ints2_ret, self.cache)
            complete_int1_uuid = _load_nested_alias_uuid(int1_ret, self.cache)
            complete_list_uuids = _load_alias_uuids(list_ret)
            complete_dict_uuids = _load_alias_uuids(dict_ret)
            complete_identity_uuid = _load_alias_uuid(identity_ret)
            complete_viz_uuid = _load_alias_uuid(viz_ret)

            # Assert that the artifacts returned by the completed run of the
            # pipeline are aliases of the artifacts created by the first failed
            # run
            self.assertEqual(ints1_uuids, complete_ints1_uuids)
            self.assertEqual(ints2_uuids, complete_ints2_uuids)
            self.assertEqual(int1_uuid, complete_int1_uuid)
            self.assertEqual(list_uuids, complete_list_uuids)
            self.assertEqual(dict_uuids, complete_dict_uuids)
            self.assertEqual(identity_uuid, complete_identity_uuid)
            self.assertEqual(viz_uuid, complete_viz_uuid)

    def test_nested_resumable_pipeline_parallel(self):
        with self.pool:
            with self.assertRaises(PipelineError) as e:
                with ParallelConfig():
                    future = self.nested_pipeline.parallel(
                        self.ints1, self.ints2, self.md1, self.int1, 'Hi',
                        fail=True)
                    future._result()

            ints1_uuids, ints2_uuids, int1_uuid, list_uuids, dict_uuids, \
                identity_uuid, viz_uuid = e.exception.uuids

            with ParallelConfig():
                future = self.nested_pipeline.parallel(
                        self.ints1, self.ints2, self.md1, self.int1, 'Hi')
                ints1_ret, ints2_ret, int1_ret, list_ret, dict_ret, \
                    identity_ret, viz_ret = future._result()

            complete_ints1_uuids = _load_nested_alias_uuids(
                ints1_ret, self.cache)
            complete_ints2_uuids = _load_nested_alias_uuids(
                ints2_ret, self.cache)
            complete_int1_uuid = _load_nested_alias_uuid(int1_ret, self.cache)
            complete_list_uuids = _load_alias_uuids(list_ret)
            complete_dict_uuids = _load_alias_uuids(dict_ret)
            complete_identity_uuid = _load_alias_uuid(identity_ret)
            complete_viz_uuid = _load_alias_uuid(viz_ret)

            # Assert that the artifacts returned by the completed run of the
            # pipeline are aliases of the artifacts created by the first failed
            # run
            self.assertEqual(ints1_uuids, complete_ints1_uuids)
            self.assertEqual(ints2_uuids, complete_ints2_uuids)
            self.assertEqual(int1_uuid, complete_int1_uuid)
            self.assertEqual(list_uuids, complete_list_uuids)
            self.assertEqual(dict_uuids, complete_dict_uuids)
            self.assertEqual(identity_uuid, complete_identity_uuid)
            self.assertEqual(viz_uuid, complete_viz_uuid)

    def test_resumable_pipeline_default_args(self):
        with self.pool:
            with self.assertRaises(PipelineError) as e:
                self.pipeline(
                    self.ints1, self.ints2, self.md1,
                    fail=True)

            ints1_uuids, ints2_uuids, int1_uuid, list_uuids, dict_uuids, \
                identity_uuid, viz_uuid = e.exception.uuids

            ints1_ret, ints2_ret, int1_ret, list_ret, dict_ret, \
                identity_ret, viz_ret = self.pipeline(
                    self.ints1, self.ints2, self.md1)

            complete_ints1_uuids = _load_alias_uuids(ints1_ret)
            complete_ints2_uuids = _load_alias_uuids(ints2_ret)
            complete_int1_uuid = _load_alias_uuid(int1_ret)
            complete_list_uuids = _load_alias_uuids(list_ret)
            complete_dict_uuids = _load_alias_uuids(dict_ret)
            complete_identity_uuid = _load_alias_uuid(identity_ret)
            complete_viz_uuid = _load_alias_uuid(viz_ret)

            # Assert that the artifacts returned by the completed run of
            # the pipeline are aliases of the artifacts created by the
            # first failed run
            self.assertEqual(ints1_uuids, complete_ints1_uuids)
            self.assertEqual(ints2_uuids, complete_ints2_uuids)
            self.assertEqual(int1_uuid, complete_int1_uuid)
            self.assertEqual(list_uuids, complete_list_uuids)
            self.assertEqual(dict_uuids, complete_dict_uuids)
            self.assertEqual(identity_uuid, complete_identity_uuid)
            self.assertEqual(viz_uuid, complete_viz_uuid)

    def test_resumable_pipeline_default_args_parallel(self):
        with self.pool:
            with self.assertRaises(PipelineError) as e:
                with ParallelConfig():
                    future = self.pipeline.parallel(
                        self.ints1, self.ints2, self.md1,
                        fail=True)
                    future._result()

            ints1_uuids, ints2_uuids, int1_uuid, list_uuids, dict_uuids, \
                identity_uuid, viz_uuid = e.exception.uuids

            with ParallelConfig():
                future = self.pipeline.parallel(
                    self.ints1, self.ints2, self.md1)
                ints1_ret, ints2_ret, int1_ret, list_ret, dict_ret, \
                    identity_ret, viz_ret = future._result()

            complete_ints1_uuids = _load_alias_uuids(ints1_ret)
            complete_ints2_uuids = _load_alias_uuids(ints2_ret)
            complete_int1_uuid = _load_alias_uuid(int1_ret)
            complete_list_uuids = _load_alias_uuids(list_ret)
            complete_dict_uuids = _load_alias_uuids(dict_ret)
            complete_identity_uuid = _load_alias_uuid(identity_ret)
            complete_viz_uuid = _load_alias_uuid(viz_ret)

            # Assert that the artifacts returned by the completed run of
            # the pipeline are aliases of the artifacts created by the
            # first failed run
            self.assertEqual(ints1_uuids, complete_ints1_uuids)
            self.assertEqual(ints2_uuids, complete_ints2_uuids)
            self.assertEqual(int1_uuid, complete_int1_uuid)
            self.assertEqual(list_uuids, complete_list_uuids)
            self.assertEqual(dict_uuids, complete_dict_uuids)
            self.assertEqual(identity_uuid, complete_identity_uuid)
            self.assertEqual(viz_uuid, complete_viz_uuid)

    def test_nested_resumable_pipeline_default_args(self):
        with self.pool:
            with self.assertRaises(PipelineError) as e:
                self.nested_pipeline(
                    self.ints1, self.ints2, self.md1, fail=True)

            ints1_uuids, ints2_uuids, int1_uuid, list_uuids, dict_uuids, \
                identity_uuid, viz_uuid = e.exception.uuids

            # We now run the not nested version. This will be able to reuse the
            # returns from varied_method
            ints1_ret, ints2_ret, int1_ret, list_ret, dict_ret, \
                identity_ret, viz_ret = self.nested_pipeline(
                    self.ints1, self.ints2, self.md1)

            complete_ints1_uuids = _load_nested_alias_uuids(
                ints1_ret, self.cache)
            complete_ints2_uuids = _load_nested_alias_uuids(
                ints2_ret, self.cache)
            complete_int1_uuid = _load_nested_alias_uuid(int1_ret, self.cache)
            complete_list_uuids = _load_alias_uuids(list_ret)
            complete_dict_uuids = _load_alias_uuids(dict_ret)
            complete_identity_uuid = _load_alias_uuid(identity_ret)
            complete_viz_uuid = _load_alias_uuid(viz_ret)

            # Assert that the artifacts returned by the completed run of the
            # pipeline are aliases of the artifacts created by the first failed
            # run
            self.assertEqual(ints1_uuids, complete_ints1_uuids)
            self.assertEqual(ints2_uuids, complete_ints2_uuids)
            self.assertEqual(int1_uuid, complete_int1_uuid)
            self.assertEqual(list_uuids, complete_list_uuids)
            self.assertEqual(dict_uuids, complete_dict_uuids)
            self.assertEqual(identity_uuid, complete_identity_uuid)
            self.assertEqual(viz_uuid, complete_viz_uuid)

    def test_nested_resumable_pipeline_parallel_default_args(self):
        with self.pool:
            with self.assertRaises(PipelineError) as e:
                with ParallelConfig():
                    future = self.nested_pipeline.parallel(
                        self.ints1, self.ints2, self.md1, fail=True)
                    future._result()

            ints1_uuids, ints2_uuids, int1_uuid, list_uuids, dict_uuids, \
                identity_uuid, viz_uuid = e.exception.uuids

            with ParallelConfig():
                future = self.nested_pipeline.parallel(
                        self.ints1, self.ints2, self.md1)
                ints1_ret, ints2_ret, int1_ret, list_ret, dict_ret, \
                    identity_ret, viz_ret = future._result()

            complete_ints1_uuids = _load_nested_alias_uuids(
                ints1_ret, self.cache)
            complete_ints2_uuids = _load_nested_alias_uuids(
                ints2_ret, self.cache)
            complete_int1_uuid = _load_nested_alias_uuid(int1_ret, self.cache)
            complete_list_uuids = _load_alias_uuids(list_ret)
            complete_dict_uuids = _load_alias_uuids(dict_ret)
            complete_identity_uuid = _load_alias_uuid(identity_ret)
            complete_viz_uuid = _load_alias_uuid(viz_ret)

            # Assert that the artifacts returned by the completed run of the
            # pipeline are aliases of the artifacts created by the first failed
            # run
            self.assertEqual(ints1_uuids, complete_ints1_uuids)
            self.assertEqual(ints2_uuids, complete_ints2_uuids)
            self.assertEqual(int1_uuid, complete_int1_uuid)
            self.assertEqual(list_uuids, complete_list_uuids)
            self.assertEqual(dict_uuids, complete_dict_uuids)
            self.assertEqual(identity_uuid, complete_identity_uuid)
            self.assertEqual(viz_uuid, complete_viz_uuid)



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------
import abc
import os
import pandas as pd
import pathlib
import tempfile
import yaml
import warnings
from zipfile import ZipFile

from dataclasses import dataclass
from datetime import timedelta
from io import BytesIO
from typing import Any, Dict, List, Optional, Set, Tuple, Union

import bibtexparser as bp
import networkx as nx

from ._checksum_validator import (
    ValidationCode, ChecksumDiff, validate_checksums
)
from .util import get_root_uuid, get_nonroot_uuid, parse_version
from ..provenance import MetadataInfo


@dataclass
class Config():
    '''
    Dataclass that stores user-selected configuration options.

    Attributes
    ----------
    perform_checksum_validation : bool
        Whether to opt in or out of checksum validation.
    parse_study_metadata : bool
        Whether to parse study metadata stored in provenance.
    recurse : bool
        Whether to recursively parse nested directories that contain artifacts.
    verbose : bool
        Whether to print status messages to stdout during processing.
    '''
    perform_checksum_validation: bool = True
    parse_study_metadata: bool = True
    recurse: bool = False
    verbose: bool = False


@dataclass
class ParserResults():
    '''
    Results generated and returned by a ParserVx.

    Attributes
    ----------
    parsed_artifact_uuids : set of str
        The uuids of the artifacts directly parsed by a parser. Does not
        include the uuids of artifact parsed from provenance. When parsing
        a single archive this is a single member set of that uuid. When
        parsing a directory, it is the set of all artifact uuids in that
        directory.
    prov_digraph : nx.Digraph
        The directed acyclic graph representation of the parsed provenance as
        an nx.DiGraph object.
    provenance_is_valid : ValidationCode
        A flag indicating the level of checksum validation.
    checksum_diff : ChecksumDiff or None
        A tuple of three dictionaries indicating the uuids of files that have
        been 1) added 2) removed or 3) changed in the archive since the
        archive was checksummed.
        None if no checksum validation was perfomed, e.g. when opted out or
        impossible because archive version did not support checksums, or when
        checksums.md5 missing from archive where it was expected.
        Interpretable only in conjunction with provenance_is_valid.
    '''
    parsed_artifact_uuids: Set[str]
    prov_digraph: nx.DiGraph
    provenance_is_valid: ValidationCode
    checksum_diff: Optional[ChecksumDiff]


class ProvNode:
    '''
    One node of a provenance DAG, describing one QIIME2 Result.
    '''

    @property
    def _uuid(self) -> str:
        return self._result_md.uuid

    @_uuid.setter
    def _uuid(self, new_uuid: str):
        '''
        ProvNode's UUID. Safe for use as getter. Prefer ProvDAG.relabel_nodes
        as a setter because it preserves alignment between ids across the dag
        and its ProvNodes.
        '''
        self._result_md.uuid = new_uuid

    @property
    def type(self) -> str:
        return self._result_md.type

    @property
    def format(self) -> Optional[str]:
        return self._result_md.format

    @property
    def archive_version(self) -> str:
        return self._archive_version

    @property
    def framework_version(self) -> str:
        return self._framework_version

    @property
    def has_provenance(self) -> bool:
        return int(self.archive_version) > 1

    @property
    def citations(self) -> Dict:
        citations = {}
        if hasattr(self, '_citations'):
            citations = self._citations.citations
        return citations

    @property
    def metadata(self) -> Optional[Dict[str, pd.DataFrame]]:
        '''
        A dict containing {parameter_name: metadata_dataframe} pairs where
        parameter_name is the registered name of the parameter the Metadata
        or MetadataColumn was passed to.

        Returns an empty dict if this action takes no Metadata or
        MetadataColumn.

        Returns None if this action has no metadata because the archive has no
        provenance, or the user opted out of metadata parsing.
        '''
        self._metadata: Optional[Dict[str, pd.DataFrame]]

        md = None
        if hasattr(self, '_metadata'):
            md = self._metadata
        return md

    @property
    def _parents(self) -> Optional[List[Dict[str, str]]]:
        '''
        A list of single-item {Type: UUID} dicts describing this
        action's inputs, including Artifacts passed as Metadata parameters.

        Returns [] if this action is an Import.

        NOTE: This property is private because it is slightly unsafe,
        reporting original node IDs that are not updated if the user renames
        nodes using the networkx API instead of ProvDAG.relabel_nodes.
        ProvDAG and its extensions should use the networkx.DiGraph itself to
        work with ancestry when possible.
        '''
        if not self.has_provenance:
            return None

        inputs = self.action._action_details.get('inputs')
        parents = []
        if inputs is not None:
            # Inputs are a list of single-item dicts
            for input in inputs:
                (name, value), = input.items()
                # value is usually a uuid, but may be a collection of uuids
                # the following are specced in qiime2/core/type/collection
                if type(value) in (set, list, tuple):
                    for i in range(len(value)):
                        # Make these unique in case the single-item dicts get
                        # merged into a single dict downstream
                        if type(value[i]) is dict:
                            unq_name, = value[i].keys()
                            v, = value[i].values()
                        else:
                            unq_name = f'{name}_{i}'
                            v = value[i]
                        parents.append({unq_name: v})
                elif value is not None:
                    parents.append({name: value})
                else:
                    # skip None-by-default optional inputs
                    pass

        return parents + self._artifacts_passed_as_md

    def __init__(
        self,
        cfg: Config,
        zf: ZipFile,
        node_fps: List[pathlib.Path]
    ):
        '''
        Constructs a ProvNode from a zipfile and the collected
        provenance-relevant filepaths for a single result within it.
        '''
        for fp in node_fps:
            if fp.name == 'VERSION':
                self._archive_version, self._framework_version = \
                    parse_version(zf)
            elif fp.name == 'metadata.yaml':
                self._result_md = _ResultMetadata(zf, str(fp))
            elif fp.name == 'action.yaml':
                self.action = _Action(zf, str(fp))
            elif fp.name == 'citations.bib':
                self._citations = _Citations(zf, str(fp))
            elif fp.name == 'checksums.md5':
                # Handled in ProvDAG
                pass

        if self.has_provenance:
            all_metadata_fps, self._artifacts_passed_as_md = \
                self._get_metadata_from_Action(self.action._action_details)
            if cfg.parse_study_metadata:
                self._metadata = self._parse_metadata(zf, all_metadata_fps)

    def _get_metadata_from_Action(
        self, action_details: Dict[str, List]
    ) -> Tuple[Dict[str, str], List[Dict[str, str]]]:
        '''
        Gathers data related to Metadata and MetadataColumn-based metadata
        files from the parsed action.yaml file.

        Captures filepath and parameter-name data for all study metadata
        files, so that these can be located for parsing, and then associated
        with the correct parameters during replay. It captures uuids for all
        artifacts passed to this action as metadata so they can be included as
        parents of this node.

        Parameters
        ----------
        action_details : dict
            The parsed dictionary of the `action` section from action.yaml.

        Returns
        -------
        tuple of (all_metadata, artifacts_as_metadata)
            Where all_metadata is a dict of
            {parameter_name: filename}.
            Where artifacts_as_metadata is a list of single-items dict of the
            structure {'artifact_passed_as_metadata': <uuid>}.

        Notes
        -----
        When Artifacts are passed as Metadata, they are captured in
        action['parameters'], rather than in action['inputs'] with the other
        Artifacts. Semantic Type data is thus not captured. This function
        returns a filler 'Type' for all UUIDs discovered here:
        'artifact_passed_as_metadata'. Because Artifacts passed (viewed) as
        Metadata retain their provenance, downstream Artifacts are linked to
        their real parent Artifact nodes with the proper Type information.
        '''
        all_metadata = dict()
        artifacts_as_metadata = []
        if (all_params := action_details.get('parameters')) is not None:
            for param in all_params:
                param_val, = param.values()
                if isinstance(param_val, MetadataInfo):
                    param_name, = param.keys()
                    md_fp = param_val.relative_fp
                    all_metadata.update({param_name: md_fp})

                    artifacts_as_metadata += [
                        {'artifact_passed_as_metadata': uuid} for uuid in
                        param_val.input_artifact_uuids
                    ]

        return all_metadata, artifacts_as_metadata

    def _parse_metadata(
        self, zf: ZipFile, metadata_fps: Dict[str, str]
    ) -> Dict[str, pd.DataFrame]:
        '''
        Parses all metadata files captured from Metadata and MetadataColumns
        (identifiable by !metadata tags) into pd.DataFrames.

        Parameters
        ----------
        zf : ZipFile
            The zipfile object of the archive.
        metadata_fps : dict
            A dict of parameter names to metadata filenames for metadata
            paramters.

        Returns
        -------
        dict
            A dict of parameter names to dataframe objects that is loaded from
            the corresponding metadata file.

            An empty dict if there is no metadata.
        '''
        if metadata_fps == {}:
            return {}

        root_uuid = get_root_uuid(zf)
        pfx = pathlib.Path(root_uuid) / 'provenance'
        if root_uuid == self._uuid:
            pfx = pfx / 'action'
        else:
            pfx = pfx / 'artifacts' / self._uuid / 'action'

        all_md = dict()
        for param_name in metadata_fps:
            filepath = str(pfx / metadata_fps[param_name])
            with zf.open(filepath) as fh:
                df = pd.read_csv(BytesIO(fh.read()), sep='\t')
                all_md[param_name] = df

        return all_md

    def __repr__(self) -> str:
        return repr(self._result_md)

    __str__ = __repr__

    def __hash__(self) -> int:
        return hash(self._uuid)

    def __eq__(self, other) -> bool:
        return (
            self.__class__ == other.__class__ and self._uuid == other._uuid
        )


class _Action:
    '''Provenance data from action.yaml for a single QIIME2 Result.'''

    @property
    def action_id(self) -> str:
        '''The UUID of the Action itself.'''
        return self._execution_details['uuid']

    @property
    def action_type(self) -> str:
        '''
        The type of Action represented e.g. Method, Pipeline, et al.
        '''
        return self._action_details['type']

    @property
    def runtime(self) -> timedelta:
        '''The elapsed run time of the Action, as a datetime object.'''
        end = self._execution_details['runtime']['end']
        start = self._execution_details['runtime']['start']
        return end - start

    @property
    def runtime_str(self) -> str:
        ''' The elapsed run time of the Action in seconds and microseconds.'''
        return self._execution_details['runtime']['duration']

    @property
    def action_name(self) -> str:
        '''
        The name of the action itself. Imports return 'import'.
        '''
        if self.action_type == 'import':
            return 'import'
        return self._action_details.get('action')

    @property
    def plugin(self) -> str:
        '''
        The plugin which executed this Action. Returns 'framework' if this is
        an import.
        '''
        if self.action_type == 'import':
            return 'framework'

        plugin = self._action_details.get('plugin')
        return plugin.replace('-', '_')

    @property
    def inputs(self) -> dict:
        '''
        Creates a dict of artifact inputs to this action.

        Returns
        -------
        dict
            A mapping of input name to the data type passed for that input
            (either uuid, list of uuid, or dict), see below for details.

        Notes
        -----
        One of three structures may be encountered when parsing this section of
        action.yaml, described below:

        case 1:

            inputs:
            - some_input_name: some_uuid
            - some_other_input_name: some_other_uuid
            (...)

        case 2:

            inputs:
            - some_input_name:
                - some_uuid
                - some_other_uuid
            (...)

        case 3 (result collection):

            inputs:
            - result_collection_name:
                - some_key: some_uuid
                - some_other_key: some_other_uuid
            (...)

            and thus is a different structure entirely.
        '''
        inputs = self._action_details.get('inputs')
        results = {}
        if inputs is not None:
            for input_ in inputs:
                nest_lvl_1 = next(iter(input_.values()))
                if type(nest_lvl_1) is list and type(nest_lvl_1[0]) is dict:
                    # result collection
                    rc = {}
                    for member in nest_lvl_1:
                        rc.update(member)

                    input_name = next(iter(input_))
                    results.update({input_name: rc})
                else:
                    # not result collection
                    results.update(input_)

        return results

    @property
    def input_result_collections(self):
        '''
        Collects all result collections passed as inputs (if any). Used for
        constructing the result collection namespace.

        Returns
        -------
        list of str
            A list of the names of the result collections passed as input.
            The names are as registered in the method registration.
        '''
        result_collection_names = []
        for key, value in self.inputs:
            if type(value) is dict:
                result_collection_names.append(key)

        return result_collection_names

    @property
    def parameters(self) -> dict:
        '''Returns a dict of parameters passed to this action.'''
        params = self._action_details.get('parameters')
        results = {}
        if params is not None:
            for item in params:
                results.update(item.items())
        return results

    @property
    def output_name(self) -> Optional[str]:
        '''
        Gets the output name of the node.

        Returns
        -------
        str or None
            The name of the output as parsed from action.yaml, or None if there
            is no output-name section.
        '''
        output_name = self._action_details.get('output-name')

        if type(output_name) is list:
            output_name = output_name[0]

        return output_name

    @property
    def result_collection_key(self) -> Optional[str]:
        '''
        Gets the result collection key if the artifact is part of a result
        collection.

        Returns
        -------
        str
            The result collection key if the artifact was output as part of
            a result collection, none otherwise.

        Notes
        -----
        We know if the artifact comes from a ResultCollection because
        outputs from a ResultCollection look like:

        output-name:
        - output
        - key
        - position/total positions
        '''
        output_name = self._action_details.get('output-name')

        if type(output_name) is not list:
            return None

        return output_name[1]

    @property
    def format(self) -> Optional[str]:
        '''Returns this action's format field if any.'''
        return self._action_details.get('format')

    @property
    def transformers(self) -> Optional[Dict]:
        '''Returns this action's transformers dictionary if any.'''
        return self._action_dict.get('transformers')

    def __init__(self, zf: ZipFile, fp: str):
        with tempfile.TemporaryDirectory() as tempdir:
            zf.extractall(tempdir)
            action_fp = os.path.join(tempdir, fp)
            with open(action_fp) as fh:
                self._action_dict = yaml.safe_load(fh)

        self._action_details = self._action_dict['action']
        self._execution_details = self._action_dict['execution']

    def __repr__(self):
        return (
            f'_Action(action_id={self.action_id}, type={self.action_type},'
            f' plugin={self.plugin}, action={self.action_name})'
        )


class _Citations:
    '''
    Citations for a single QIIME2 Result, as a dict of citation dicts keyed
    on the citation's bibtex ID.
    '''
    def __init__(self, zf: ZipFile, fp: str):
        bib_db = bp.loads(zf.read(fp))
        self.citations = bib_db.get_entry_dict()

    def __repr__(self):
        keys = list(self.citations.keys())
        return f'Citations({keys})'


class _ResultMetadata:
    '''Basic metadata about a single QIIME2 Result from metadata.yaml.'''
    def __init__(self, zf: ZipFile, md_fp: str):
        _md_dict = yaml.safe_load(zf.read(md_fp))
        self.uuid = _md_dict['uuid']
        self.type = _md_dict['type']
        self.format = _md_dict['format']

    def __repr__(self):
        return (
            f'UUID:\t\t{self.uuid}\n'
            f'Type:\t\t{self.type}\n'
            f'Data Format:\t{self.format}'
        )


class Parser(metaclass=abc.ABCMeta):
    accepted_data_types: str

    @classmethod
    @abc.abstractmethod
    def get_parser(cls, artifact_data: Any) -> 'Parser':
        '''
        Return the appropriate Parser if this Parser type can handle the data
        passed in.

        Should raise an appropriate exception if this Parser cannot handle the
        data.
        '''

    @abc.abstractmethod
    def parse_prov(self, cfg: Config, data: Any) -> ParserResults:
        '''
        Parse provenance to return a ParserResults.
        '''


class ArchiveParser(Parser):
    accepted_data_types = 'a path to a file (a string) or a file-like object'

    @classmethod
    def get_parser(cls, artifact: Union[str, pathlib.PosixPath]) -> Parser:
        '''
        Returns the correct archive format parser for a zip archive.

        Parameters
        ----------
        artifact_data : str or pathlib.PosixPath
            A path to a zipped archive.

        Returns
        -------
        Parser
            An ArchiveParser object for the version of the artifact. One of
            ParserV[0-6].
        '''
        if type(artifact) is pathlib.PosixPath:
            artifact = str(artifact)
        if type(artifact) is not str:
            raise TypeError(
                'ArchiveParser expects a string or pathlib.PosixPath path to '
                f'an archive, not an object of type {str(type(artifact))}.'
            )
        if os.path.isdir(artifact):
            raise ValueError('ArchiveParser expects a file, not a directory.')

        try:
            with ZipFile(artifact, 'r') as zf:
                archive_version, _ = parse_version(zf)
            return FORMAT_REGISTRY[archive_version]()
        except KeyError as e:
            raise KeyError(
                f'While trying to parse artifact {artifact}, '
                'a corresponding parser was not found for archive version '
                f'{archive_version}: {str(e)}.'
            )

    def parse_prov(cls, cfg: Config, data: Any) -> ParserResults:
        raise NotImplementedError(
            'Use a subclass that usefully defines parse_prov for some format.'
        )


class ParserV0(ArchiveParser):
    '''
    Parser for V0 archives. V0 archives have no ancestral provenance.
    '''
    # These are files we expect will be present in every QIIME2 archive with
    # this format. "Optional" filenames (like Metadata, which may or may
    # not be present in an archive) should not be included here.
    expected_files_root_only = tuple()
    expected_files_all_nodes = ('metadata.yaml', 'VERSION')

    def parse_prov(self, cfg: Config, archive: str) -> ParserResults:
        '''
        Parses an artifact's provenance into a directed acyclic graph.

        In the case of v0 archives, the only provenance information is that
        which is attached to the artifact itself; information about ancestor
        nodes does not exist. The parsed dag contains only a single node.

        In the case of v1 archives, ancestor nodes do exist in the
        archive. However, because the corresponding action.yaml does not track
        output names, when two outputs share the same semantic type, it is not
        possible to untangle provenance. Instead of wrangling with this and
        in consideration of the expected rarity of v1 archives, it was decided
        to treat v1 archives as v0 archives.

        Parameters
        ----------
        cfg : Config
            A dataclass that stores four boolean flags: whether to perform
            checksum validation, whether to parse study metadata, whether to
            recursively parse nested directories, and whether to enable verbose
            mode.
        archive : str
            A path to the artifact to be parsed.

        Returns
        -------
        ParserResults
            A dataclass that stores the parsed artifact uuids, the parsed
            networkx graph, the provenance-is-valid flag, and the
            checksum diff.
        '''
        with ZipFile(archive) as zf:
            if cfg.perform_checksum_validation:
                provenance_is_valid, checksum_diff = \
                    self._validate_checksums(zf)
            else:
                provenance_is_valid = ValidationCode.VALIDATION_OPTOUT
                checksum_diff = None

            root_uuid = get_root_uuid(zf)
            warnings.warn(
                f'Artifact {root_uuid} was created prior to provenance '
                'tracking. Provenance data will be incomplete.',
                UserWarning
            )

            exp_node_fps = []
            for fp in self.expected_files_all_nodes:
                exp_node_fps.append(pathlib.Path(root_uuid) / fp)

            prov_fps = self._get_provenance_fps(zf)
            self._assert_expected_files_present(
                zf, exp_node_fps, prov_fps
            )
            # we have confirmed that all expected fps for this node exist
            node_fps = exp_node_fps

            nodes = {}
            nodes[root_uuid] = ProvNode(cfg, zf, node_fps)
            graph = self._digraph_from_archive_contents(nodes)

        return ParserResults(
            {root_uuid},
            graph,
            provenance_is_valid,
            checksum_diff
        )

    def _parse_root_md(self, zf: ZipFile, root_uuid: str) -> _ResultMetadata:
        '''
        Parses the root metadata file of an archive for its uuid, semantic
        type, and format.

        Parameters
        ----------
        zf : ZipFile
            A zipfile object of a v0 artifact.
        root_uuid : str
            The uuid of the root node. Because this operates on a v0 archive,
            the root node is the only node.

        Returns
        -------
        _ResultMetadata
            An object representing the information stored in a metadata.yaml
            file, namely the uuid, type, and format fields.
        '''
        root_md_fp = os.path.join(root_uuid, 'metadata.yaml')
        if root_md_fp not in zf.namelist():
            raise ValueError(
                'Malformed Archive: root metadata.yaml file '
                f'misplaced or nonexistent in {zf.filename}'
            )
        return _ResultMetadata(zf, root_md_fp)

    def _validate_checksums(
            self, zf: ZipFile
    ) -> Tuple[ValidationCode, Optional[ChecksumDiff]]:
        '''
        Return the ValidationCode and ChecksumDiff for an archive. Because
        checksums were not introduced, until ArchiveFormat version 5,
        uses the PREDATES_CHECKSUMS flag and returns None to indicate that
        checksum diffing was not performed.

        Parameters
        ----------
        zf : ZipFile
            The zipfile object representing the archive. Ignored here but
            needed in signature for inheritance.

        Returns
        -------
        tuple of (ValidationCode, None)
            The validation code and None to indicate missing ChecksumDiff.
        '''
        return (ValidationCode.PREDATES_CHECKSUMS, None)

    def _digraph_from_archive_contents(
        self, archive_contents: Dict[str, 'ProvNode']
    ) -> nx.DiGraph:
        '''
        Builds a networkx.DiGraph from a {UUID: ProvNode} dictionary.

        1. Create an empty nx.digraph.
        2. Gather nodes and their required attributes and add them to the
           DiGraph.
        3. Add edges to graph (including all !no-provenance nodes)
        4. Create guaranteed node attributes for these no-provenance nodes,
           which wouldn't otherwise have them.

        Parameters
        ----------
        archive_contents : dict of {str to ProvNode}
            A dictionary of node uuids to their representative ProvNode
            objects.

        Returns
        -------
        nx.DiGraph
            The directed, acyclic graph representation of the provenance of
            the archive. Edge directionality is from parent to child. Parents
            may have multiple children and children may have multiple parents.
        '''
        dag = nx.DiGraph()
        nodes = []
        for node_uuid, node in archive_contents.items():
            node_info = {
                'node_data': node,
                'has_provenance': node.has_provenance
            }
            nodes.append((node_uuid,  node_info))
        dag.add_nodes_from(nodes)

        edges = []
        for node_uuid, attrs in dag.nodes(data=True):
            if parents := attrs['node_data']._parents:
                for parent in parents:
                    parent_uuid, = parent.values()
                    edges.append((parent_uuid, node_uuid))
        dag.add_edges_from(edges)

        return dag

    def _get_provenance_fps(self, zf: ZipFile) -> List[pathlib.Path]:
        '''
        Collect filepaths of all provenance-relevant files in an archive.
        Relevant is defined by `self.expected_files_all_nodes` and
        `self.expected_files_root_only` (which is empty).

        Parameters
        ----------
        zf : ZipFile
            The zipfile object of the archive.

        Returns
        -------
        list of pathlib.Path
            Filepaths relative to root of zipfile for each file of interest.
        '''
        fps = []
        for fp in zf.namelist():
            for expected_filename in self.expected_files_all_nodes:
                if expected_filename in fp:
                    fps.append(pathlib.Path(fp))

        return fps

    def _assert_expected_files_present(
            self,
            zf: ZipFile,
            expected_node_fps: List[pathlib.Path],
            prov_fps: List[pathlib.Path],
    ):
        '''
        Makes sure that all expected files for a given node are present in an
        archive. Raises a ValueError if not.

        Parameters
        ----------
        zf : ZipFile
            The zipfile object representing an archive.
        expected_node_fps : list of pathlib.Path
            The filepaths that are expected to be present in the zipfile for
            some node.
        prov_fps : list of pathlib.Path
            All provenance-relevant filepaths in the archive.

        Raises
        ------
        ValueError
            If there are expected provenance-relevant files missing from
            a node.
        '''
        error_contents = 'Malformed Archive: '
        root_uuid = get_root_uuid(zf)
        for fp in expected_node_fps:
            if fp not in prov_fps:
                node_uuid = get_nonroot_uuid(fp)
                error_contents += (
                    f'{fp.name} file for node {node_uuid} '
                    f'misplaced or nonexistent in {zf.filename}.\n'
                )
                error_contents += (
                    f'Archive {root_uuid} may be corrupt '
                    'or provenance may be false.'
                )
                raise ValueError(error_contents)


class ParserV1(ParserV0):
    '''
    Parser for V1 archives. Although action.yaml was introduced for this
    archive version, we are pretending that it was introduced in V2 because of
    difficulties untangling provenance without output names. V1 archives are
    treated as having no provenance, like V0 archives.
    '''
    expected_files_root_only = ParserV0.expected_files_root_only
    expected_files_all_nodes = ParserV0.expected_files_all_nodes


class ParserV2(ParserV1):
    '''
    Parser for V2 archives. Introduces action/action.yaml to provenance.
    Directory structure identical to V1, action.yaml changes to support
    Pipelines.
    '''
    expected_files_root_only = ParserV1.expected_files_root_only
    expected_files_all_nodes = (
        *ParserV1.expected_files_all_nodes, 'action/action.yaml'
    )

    def parse_prov(self, cfg: Config, archive: str) -> ParserResults:
        '''
        Parses an artifact's provenance into a directed acyclic graph.

        For each artifact in provenance, gathers all corresponding
        provenance-relevant files and constructs a ProvNode. Once all
        ProvNodes are constructed, creates the provenance graph.

        Parameters
        ----------
        cfg : Config
            A dataclass that stores four boolean flags: whether to perform
            checksum validation, whether to parse study metadata, whether to
            recursively parse nested directories, and whether to enable verbose
            mode.
        archive_data : str
            A path to the artifact to be parsed.

        Returns
        -------
        ParserResults
            A dataclass that stores the parsed artifact uuids, the parsed
            networkx graph, the provenance-is-valid flag, and the
            checksum diff.
        '''
        with ZipFile(archive) as zf:
            if cfg.perform_checksum_validation:
                provenance_is_valid, checksum_diff = \
                    self._validate_checksums(zf)
            else:
                provenance_is_valid = ValidationCode.VALIDATION_OPTOUT
                checksum_diff = None

            prov_fps = self._get_provenance_fps(zf)
            root_uuid = get_root_uuid(zf)

            # make a provnode for each UUID
            archive_contents = {}
            for fp in prov_fps:
                exp_node_fps = []
                if 'artifacts' not in fp.parts:
                    node_uuid = root_uuid
                    prefix = pathlib.Path(node_uuid) / 'provenance'
                    root_only_expected_fps = []
                    for exp_filename in self.expected_files_root_only:
                        root_only_expected_fps.append(
                            pathlib.Path(node_uuid) / exp_filename
                        )
                    exp_node_fps += root_only_expected_fps
                else:
                    node_uuid = get_nonroot_uuid(fp)
                    # /root-uuid/provenance/artifacts/node-uuid
                    prefix = pathlib.Path(*fp.parts[0:4])

                if node_uuid in archive_contents:
                    continue

                for expected_file in self.expected_files_all_nodes:
                    exp_node_fps.append(prefix / expected_file)

                self._assert_expected_files_present(
                    zf, exp_node_fps, prov_fps
                )
                # we have confirmed that all expected fps for this node exist
                node_fps = exp_node_fps

                archive_contents[node_uuid] = ProvNode(cfg, zf, node_fps)

        graph = self._digraph_from_archive_contents(archive_contents)

        return ParserResults(
            {root_uuid},
            graph,
            provenance_is_valid,
            checksum_diff
        )

    def _get_provenance_fps(self, zf: ZipFile) -> List[pathlib.Path]:
        '''
        Collect filepaths of all provenance-relevant files in an archive.
        Relevant is defined by `self.expected_files_all_nodes` and
        `self.expected_files_root_only`.

        Parameters
        ----------
        zf : ZipFile
            The zipfile object of the archive.

        Returns
        -------
        list of pathlib.Path
            Filepaths relative to root of zipfile for each file of interest.
        '''
        fps = []
        for fp in zf.namelist():
            for expected_filename in self.expected_files_all_nodes:
                if 'provenance' in fp and expected_filename in fp:
                    fps.append(pathlib.Path(fp))

        root_uuid = get_root_uuid(zf)
        for expected_filename in self.expected_files_root_only:
            fps.append(pathlib.Path(root_uuid) / expected_filename)

        return fps


class ParserV3(ParserV2):
    '''
    Parser for V3 archives. Directory structure identical to V1 & V2,
    action.yaml now supports variadic inputs, so !set tags in action.yaml.
    '''
    expected_files_root_only = ParserV2.expected_files_root_only
    expected_files_all_nodes = ParserV2.expected_files_all_nodes


class ParserV4(ParserV3):
    '''
    Parser for V4 archives. Adds citations to directory structure, changes to
    action.yaml including transformers.
    '''
    expected_files_root_only = ParserV3.expected_files_root_only
    expected_files_all_nodes = (
        *ParserV3.expected_files_all_nodes, 'citations.bib'
    )


class ParserV5(ParserV4):
    '''
    Parser for V5 archives. Adds checksum validation with checksums.md5.
    '''
    expected_files_root_only = ('checksums.md5', )
    expected_files_all_nodes = ParserV4.expected_files_all_nodes

    def _validate_checksums(
            self, zf: ZipFile
    ) -> Tuple[ValidationCode, Optional[ChecksumDiff]]:
        '''
        Checksum support added for v5, so perform checksum validation.

        Parameters
        ----------
        zf : ZipFile
            The zipfile object representation of the parsed archive.

        Returns
        -------
        tuple of (ValidationCode, ChecksumDiff or None)
            Where ValidationCode is one of valid, invalid, predates checksums,
            optout.
            Where ChecksumDiff contains filepaths of all changed, added, and
            removed files since last checksumming.
            If checksums.md5 is missing from archive the archive, an invalid
            code is returned and a ChecksumDiff of None is returned.

        Notes
        -----
        Because a ChecksumDiff of None here has a different interpetation
        than in pre-V5 archive parsers, the ChecksumDiff should only be
        intepreted in conjuction with the ValidationCode.
        '''
        return validate_checksums(zf)


class ParserV6(ParserV5):
    '''
    Parser for V6 archives. Adds support for output collections, adds
    execution_context field to action.yaml.
    '''
    expected_files_root_only = ParserV5.expected_files_root_only
    expected_files_all_nodes = ParserV5.expected_files_all_nodes


FORMAT_REGISTRY = {
    # NOTE: update for new format versions in qiime2.core.archive.Archiver
    '0': ParserV0,
    '1': ParserV1,
    '2': ParserV2,
    '3': ParserV3,
    '4': ParserV4,
    '5': ParserV5,
    '6': ParserV6
}
